# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dr14meter.tagger import Tagger
import multiprocessing
import os
import tempfile
import fileinput
import time
import sys

from dr14meter import dr14_global
from dr14meter import dr14_config as config
from dr14meter import audio_analysis as aa
from dr14meter.dynamic_range_meter import DynamicRangeMeter, TextTable, BBcodeTable, HtmlTable, MediaWikiTable
from dr14meter.out_messages import print_msg


def scan_files_list(input_file, options, out_dir):
    if out_dir is None:
        out_dir = tempfile.gettempdir()
    else:
        out_dir = os.path.abspath(out_dir)

    a = time.time()

    if input_file is None:
        input_file = '-'

    files_list = []

    for line in fileinput.input(input_file):
        files_list.append(os.path.abspath(line.rstrip()))

    dr = DynamicRangeMeter()
    dr.write_to_local_db(config.db_is_enabled())

    r = dr.scan_mp(files_list=files_list, thread_cnt=get_thread_cnt())

    if r == 0:
        success = False
    else:
        write_results(dr, options, out_dir, "")

    if options.tag:
        tagger = Tagger()
        tagger.write_dr_tags(dr)

        success = True

    clock = time.time() - a
    return success, clock, r


def scan_dir_list(subdirlist, options, out_dir):
    a = time.time()

    success = False
    r = 0

    for cur_dir in subdirlist:
        dr = DynamicRangeMeter()
        dr.write_to_local_db(config.db_is_enabled())

        print_msg("\n------------------------------------------------------------ ")
        print_msg("> Scan Dir: %s \n" % cur_dir)

        if options.disable_multithread:
            r = dr.scan_dir(cur_dir)
        else:
            cpu = get_thread_cnt()
            r = dr.scan_mp(cur_dir, cpu)

        if options.tag:
            tagger = Tagger()
            tagger.write_dr_tags(dr)

        if r == 0:
            continue
        else:
            success = True

        write_results(dr, options, out_dir, cur_dir)

    clock = time.time() - a

    return success, clock, r


def get_thread_cnt():
    cpu = multiprocessing.cpu_count()
    cpu = max(2, int(round(cpu / 2)))
    return cpu


def list_rec_dirs(basedir, subdirlist=None):
    if subdirlist is None:
        subdirlist = [basedir]

    for item in os.listdir(basedir):
        item = os.path.join(basedir, item)
        if os.path.isdir(item):
            item = os.path.abspath(item)
            # print_msg( item )
            subdirlist.append(item)
            list_rec_dirs(item, subdirlist)

    return subdirlist


def write_results(dr, options, out_dir, cur_dir):
    table_format = not options.basic_table

    if out_dir is None:
        full_out_dir = os.path.join(cur_dir)
    else:
        full_out_dir = out_dir

    print_msg("DR = " + str(dr.dr14))

    if not (os.access(full_out_dir, os.W_OK)):
        full_out_dir = tempfile.gettempdir()
        print_msg(
            "--------------------------------------------------------------- ")
        print_msg("- ATTENTION !")
        print_msg(
            "- You do not have the write permission for the directory: %s " % full_out_dir)
        print_msg(
            "- The result files will be written in the tmp dir: %s " % full_out_dir)
        print_msg(
            "--------------------------------------------------------------- ")

    if options.print_std_out:
        dr.fwrite_dr("", TextTable(), table_format, True)

    if options.turn_off_out:
        return

    all_tables = False
    if 'a' in options.out_tables:
        all_tables = True

    tables_list = {
        'b': ["dr14_bbcode.txt", BBcodeTable()],
        't': ["dr14-DR"+str(dr.dr14)+".txt",TextTable()],
        'h': ["dr14.html", HtmlTable()],
        'w': ["dr14_mediawiki.txt", MediaWikiTable()]
    }

    out_list = ""

    dr.write_to_local_database()

    for code in tables_list.keys():
        if code in options.out_tables or all_tables:
            dr.fwrite_dr(os.path.join(full_out_dir, tables_list[code][0]), tables_list[code][1], table_format,
                         append=options.append, dr_database=options.dr_database)
            out_list += " %s " % tables_list[code][0]

    print_msg("")
    print_msg("- The full result has been written in the files: %s" % out_list)
    print_msg("- located in the directory: ")
    print_msg(full_out_dir)
    print_msg("")


def test_path_validity(path):
    if path in ["/", ""]:
        return False
    if os.access(path, os.W_OK):
        return True
    else:
        (h, t) = os.path.split(path)
        return test_path_validity(h)


def run_analysis_opt(options, path_name):
    flag = False

    if options.compress:
        if not dr14_global.test_compress_modules():
            sys.exit(1)

        print_msg("Start compressor:")
        comp = aa.AudioCompressor()
        comp.setCompressionModality(options.compress)
        comp.compute_track(path_name)
        flag = True

    if options.spectrogram:
        if not dr14_global.test_matplotlib_modules("Spectrogram"):
            sys.exit(1)

        print_msg("Start spectrogram:")
        spectr = aa.AudioSpectrogram()
        spectr.compute_track(path_name)
        flag = True

    if options.plot_track:
        if not dr14_global.test_matplotlib_modules("Plot track"):
            sys.exit(1)

        print_msg("Start Plot Track:")
        spectr = aa.AudioPlotTrack()
        spectr.compute_track(path_name)
        flag = True

    if options.plot_track_dst:
        if not dr14_global.test_matplotlib_modules("Plot track dst"):
            sys.exit(1)

        print_msg("Start Plot Track:")
        spectr = aa.AudioPlotTrackDistribution()
        spectr.compute_track(path_name)
        flag = True

    if options.histogram:
        if not dr14_global.test_hist_modules():
            sys.exit(1)

        print_msg("Start histogram:")
        hist = aa.AudioDrHistogram()
        hist.compute_track(path_name)
        flag = True

    if options.lev_histogram:
        if not dr14_global.test_hist_modules():
            sys.exit(1)

        print_msg("Start level histogram:")
        hist = aa.AudioLevelHistogram()
        hist.compute_track(path_name)
        flag = True

    if options.dynamic_vivacity:
        if not dr14_global.test_hist_modules():
            sys.exit(1)

        print_msg("Start Dynamic vivacity:")
        viva = aa.AudioDynVivacity()
        viva.compute_track(path_name)
        flag = True

    return flag



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from dr14meter import dr14_global
from mutagen.mp3 import MP3
from mutagen.flac import FLAC
from mutagen.oggvorbis import OggVorbis
from mutagen.oggopus import OggOpus
from mutagen.monkeysaudio import MonkeysAudio
from mutagen.mp4 import MP4
from mutagen.id3 import ID3, TXXX
from dr14meter.audio_file_reader import *


class Tagger:

    def __init__(self):
        self.formats = ['.flac', '.mp3', '.ogg', '.opus', '.mp4',
                        '.m4a', '.wav', '.ape', '.ac3', '.wma']
        self.dir_name = ''
        self._ext = -1
    
    def write_dr_tags(self, dr):

        if not dr14_global.test_mutagen("Tagging"):
            sys.exit(1)

        self.dir_name = dr.dir_name
        
        for item in dr.res_list:
            self.read_track_new(item)

    def get_file_ext_code(self):
        return self._ext

    def read_track_new(self, item):

        (f, ext) = os.path.splitext(item['file_name'])
        ext = ext.lower()

        if ext not in self.formats:
            return False

        audio = None
        filename = self.dir_name + os.sep + item['file_name']

        if ext == '.mp3':
            audio = MP3(filename)
            audio.add_tags
        elif ext == '.flac':
            audio = FLAC(filename)
        elif ext == '.ogg':
            audio = OggVorbis(filename)
        elif ext == '.opus':
            audio = OggOpus(filename)
        elif ext in ['.mp4', '.m4a']:
            audio = MP4(filename)
        elif ext == '.wav':
            raise Exception("Tagging .wav files not supported")
        elif ext == '.ape':
            audio = MonkeysAudio(filename)
        elif ext == '.ac3':
            raise Exception("Tagging .ac3 files not supported")
        elif ext == '.wma':
            raise Exception("Tagging .wma files not supported")
        else:
            return False

        if isinstance(audio, MP3):
            audio.tags.add(TXXX(encoding=3, desc=u"DR", text=[str(item["dr14"])]))
        else:
            audio["DR"] = [str(item["dr14"])]
        
        audio.save()
        
        self._ext = self.formats.index(ext)

        return True


# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys


def float_formatter(el):
    if abs(el) >= 1.0:
        return "%.2f" % el
    else:
        return "%.2E" % el


def default_formatter(el):
    if sys.version_info[0] == 2:
        return unicode(el)
    else:
        return str(el)


def string_formatter(el):
    if sys.version_info[0] == 2:
        return unicode(el)
    else:
        return str(el)


class Table:

    def __init__(self):
        self.__float_format = "%.2f"
        self.__col_cnt = 5
        self.__ini_txt = ""
        self.__txt = ""
        self.__formatter = {}

        self.add_formatter(float, float_formatter)
        self.add_formatter(str, string_formatter)
        if sys.version_info[0] == 2:
            self.add_formatter(unicode, string_formatter)

    def _get_txt(self):
        return self.__txt

    def _set_txt(self, txt):
        self.__txt = txt

    def _append_txt(self, txt):
        self.__txt += txt

    def init_txt(self, txt=""):
        self.__ini_txt = txt

    def get_init_txt(self):
        return self.__ini_txt

    def new_table(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def end_table(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def write_table(self):
        return self.__ini_txt + self._get_txt()

    def nl(self):
        if sys.platform.startswith('win'):
            return '\n\r'
        else:
            return '\n'

    def add_formatter(self, _type, formatter):
        self.__formatter[_type] = formatter

    def format_element(self, el):
        return self.__formatter.get(type(el), default_formatter)(el)

    def append_row(self, row_el, cell_type='d'):

        if cell_type == 'd':
            n_cell = self.new_cell
            e_cell = self.end_cell
        elif cell_type == 'h':
            n_cell = self.new_hcell
            e_cell = self.end_hcell

        self.new_row()

        for i in row_el:
            n_cell()
            self.add_value(i)
            e_cell()

        self.end_row()

    def get_col_cnt(self):
        return self.__col_cnt

    def set_col_cnt(self, col_cnt):
        self.__col_cnt = col_cnt

    col_cnt = property(get_col_cnt, set_col_cnt)

    def append_separator_line(self):
        self._append_txt(self.format_element(""))

    def append_closing_line(self):
        self._append_txt(self.format_element(""))

    def append_empty_line(self):
        self.append_row([""] * self.col_cnt)

    def add_title(self, title):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def add_value(self, val):
        self._append_txt(self.format_element(val))

    def new_head(self):
        self._append_txt(self.format_element(""))

    def end_head(self):
        self._append_txt(self.format_element(""))

    def new_tbody(self):
        self._append_txt(self.format_element(""))

    def end_tbody(self):
        self._append_txt(self.format_element(""))

    def new_foot(self):
        self._append_txt(self.format_element(""))

    def end_foot(self):
        self._append_txt(self.format_element(""))

    def new_row(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def end_row(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def new_cell(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def end_cell(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def new_hcell(self):
        return self.new_cell()

    def end_hcell(self):
        return self.end_cell()

    def new_bold(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def end_bold(self):
        NotImplementedError(
            " %s : is virtual and must be overridden." % sys._getframe().f_code.co_name)


class TextTable (Table):

    def append_separator_line(self):
        self.append_row(
            ["----------------------------------------------------------------------------------------------"])

    def append_closing_line(self):
        self.append_row(
            ["=============================================================================================="])

    def append_empty_line(self):
        self.append_row([""])

    def add_title(self, title):
        self._append_txt(title + self.nl())

    def new_table(self):
        self._set_txt("")

    def end_table(self):
        self._append_txt(self.nl())

    def new_row(self):
        self._append_txt("")

    def end_row(self):
        self._append_txt(self.nl())

    def new_cell(self):
        self._append_txt("")

    def end_cell(self):
        self._append_txt("\t")

    def new_bold(self):
        self._append_txt("")

    def end_bold(self):
        self._append_txt("")


class BBcodeTable (Table):

    def append_separator_line(self):
        self.append_row(["------------"] * self.col_cnt)

    def append_closing_line(self):
        self.append_row(["============"] * self.col_cnt)

    def add_title(self, title):
        self._append_txt(self.nl() + "[tr]" + self.nl() + " [td  colspan=%d] " %
                         self.col_cnt + title + " [/td] " + self.nl() + "[/tr]" + self.nl())

    def new_table(self):
        self._set_txt("")
        self._append_txt('[table]' + self.nl())

    def end_table(self):
        self._append_txt(self.nl() + '[/table]' + self.nl())

    def new_row(self):
        self._append_txt(self.nl() + '[tr]' + self.nl())

    def end_row(self):
        self._append_txt(self.nl() + '[/tr]' + self.nl())

    def new_cell(self):
        self._append_txt(' [td]')

    def end_cell(self):
        self._append_txt('[/td]')

    def new_bold(self):
        self._append_txt('[b]')

    def end_bold(self):
        self._append_txt('[/b]')


class HtmlTable (Table):

    def add_title(self, title):
        self._append_txt(self.nl() + "<tr>" + self.nl() + " <th colspan=\"%d\" > " %
                         self.col_cnt + title + "</th>" + self.nl() + "</tr>" + self.nl())

    def new_table(self):
        self._set_txt("")
        self._append_txt("<table>" + self.nl())

    def end_table(self):
        self._append_txt(self.nl() + "</table>" + self.nl())

    def new_head(self):
        self._append_txt(self.nl() + "<thead>" + self.nl())

    def end_head(self):
        self._append_txt(self.nl() + "</thead>" + self.nl())

    def new_tbody(self):
        self._append_txt(self.nl() + "<tbody>" + self.nl())

    def end_tbody(self):
        self._append_txt(self.nl() + "</tbody>" + self.nl())

    def new_foot(self):
        self._append_txt(self.nl() + "<tfoot>" + self.nl())

    def end_foot(self):
        self._append_txt(self.nl() + "</tfoot>" + self.nl())

    def new_row(self):
        self._append_txt(self.nl() + "<tr>" + self.nl())

    def end_row(self):
        self._append_txt(self.nl() + "</tr>" + self.nl())

    def new_cell(self):
        self._append_txt(' <td>')

    def end_cell(self):
        self._append_txt('</td>')

    def new_hcell(self):
        self._append_txt(' <th>')

    def end_hcell(self):
        self._append_txt('</th>')

    def new_bold(self):
        self._append_txt('<b>')

    def end_bold(self):
        self._append_txt('</b>')


class MediaWikiTable (Table):

    def add_title(self, title):
        self._append_txt("|-" + self.nl() + "!align=\"left\" colspan=\"%d\" | " %
                         self.col_cnt + title + self.nl())

    def new_table(self):
        self._set_txt("")
        self._append_txt("{| " + self.nl())

    def end_table(self):
        self._append_txt("|}" + self.nl())

    def new_row(self):
        self._append_txt("|-" + self.nl())

    def end_row(self):
        self._append_txt(self.nl())

    def new_cell(self):
        self._append_txt('||')

    def end_cell(self):
        self._append_txt("")

    def new_bold(self):
        self._append_txt("\'\'\'")

    def end_bold(self):
        self._append_txt("\'\'\'")


class row:

    def __init__(self):
        self.row = []
        self.cursor = 0
        self.inds = []
        self.rclass = "b"
        self.type = ""
        self.cell_type = []

    def set_type(self, t):
        self.type = t

    def set_rclass(self, c):
        self.rclass = c

    @property
    def set_row(self):
        self.set_type("r")
        return self.type

    @property
    def set_head(self):
        self.set_rclass("h")
        return self.rclass

    @property
    def set_body(self):
        self.set_rclass("b")
        return self.rclass

    @property
    def set_foot(self):
        self.set_rclass("f")
        return self.rclass

    @property
    def set_title(self):
        self.set_type("t")
        return self.type

    @property
    def set_separator_line(self):
        self.set_type("sl")
        return self.type

    @property
    def set_closing_line(self):
        self.set_type("cl")
        return self.type

    @property
    def is_row(self):
        if self.type == "r":
            return True
        else:
            return False

    @property
    def is_head(self):
        if self.rclass == "h":
            return True
        else:
            return False

    @property
    def is_body(self):
        if self.rclass == "b":
            return True
        else:
            return False

    @property
    def is_foot(self):
        if self.rclass == "f":
            return True
        else:
            return False

    @property
    def is_title(self):
        if self.type == "t":
            return True
        else:
            return False

    @property
    def is_separator_line(self):
        if self.type == "sl":
            return True
        else:
            return False

    @property
    def is_closing_line(self):
        if self.type == "cl":
            return True
        else:
            return False


class ExtendedTextTable (Table):

    def __init__(self):
        Table.__init__(self)
        self._cols_sz = [0] * Table.get_col_cnt(self)
        self._rows = []
        self._bold_state = False
        self._rclass_state = "b"

    def get_col_cnt(self):
        return Table.get_col_cnt(self)

    def set_col_cnt(self, col_cnt):
        Table.set_col_cnt(self, col_cnt)

        if len(self._cols_sz) < col_cnt:
            self._cols_sz = self.col_sz[:col_cnt]
        elif len(self._cols_sz) > col_cnt:
            for n in range(len(self._cols_sz), col_cnt):
                self._cols_sz.append(0)

    col_cnt = property(get_col_cnt, set_col_cnt)

    def _eval_row_len(self):
        l = sum(self._cols_sz)
        l = l + len(self._cols_sz) * 3
        return l

    def _update_col_sz(self):
        r = self._rows[-1]

        if r.is_row and len(r.row) == self.col_cnt:
            for c, i in zip(r.row, range(self.col_cnt)):
                if len(c) > self._cols_sz[i]:
                    self._cols_sz[i] = len(c)
        elif r.is_title:
            d = self._eval_row_len() - len(r.row[0])
            if d > 0:
                c = 0
                while d > 0:
                    self._cols_sz[c] += 1
                    d -= 1
                    c = (c + 1) % len(self._cols_sz)
        elif r.is_separator_line:
            pass
        elif r.is_closing_line:
            pass
        else:
            raise Exception("%s : Row model: Not Allowed " %
                            sys._getframe().f_code.co_name)

    def _write_title(self, r):
        txt = " "
        txt += r.row[0]
        txt += self.nl()
        self._append_txt(txt)

    def _write_row(self, r):
        txt = " "
        for cell, i in zip(r.row, range(len(r.row))):
            t_txt = " "
            t_txt += cell
            a = self._cols_sz[i] - len(cell)
            if a < 0:
                a = 0
            t_txt += " " * (a + 1)
            txt += t_txt

        txt += self.nl()
        self._append_txt(txt)

    def _write_separator_line(self, r):
        l = self._eval_row_len()
        txt = " "
        txt += "=" * (l - 2)
        txt += self.nl()
        self._append_txt(txt)

    def _write_closing_line(self, r):
        self._write_separator_line(r)

    def write_table(self):

        for r in self._rows:
            if r.is_title:
                self._write_title(r)
            elif r.is_row:
                self._write_row(r)
            elif r.is_separator_line:
                self._write_separator_line(r)
            elif r.is_closing_line:
                self._write_closing_line(r)
            else:
                raise Exception("%s : Row model: Not Allowed " %
                                sys._getframe().f_code.co_name)

        return self.get_init_txt() + self._get_txt()

    def new_table(self):
        self._cols_sz = [0] * self.col_cnt
        self._rows = []

    def end_table(self):
        pass

    def append_separator_line(self):
        r = row()
        r.set_separator_line
        r.set_rclass(self._rclass_state)

        self._rows.append(r)

    def append_closing_line(self):
        r = row()
        r.set_closing_line
        r.set_rclass(self._rclass_state)

        self._rows.append(r)

    def append_empty_line(self):
        self.append_row([""] * self.col_cnt)
        self._update_col_sz()

    def add_title(self, title):
        r = row()
        r.set_title
        r.set_rclass(self._rclass_state)
        r.row.append(title)
        self._rows.append(r)
        self._update_col_sz()

    def new_row(self):
        r = row()
        r.set_row
        r.set_rclass(self._rclass_state)
        r.cursor = 0

        self._rows.append(r)

    def end_row(self):
        self._update_col_sz()

    def new_cell(self):
        self._rows[-1].inds.append(self._rows[-1].cursor)
        self._rows[-1].row.append("")
        self._rows[-1].cell_type.append("c")

    def end_cell(self):
        self._rows[-1].cursor += 1

    def add_value(self, val):
        c = self._rows[-1].cursor
        self._rows[-1].row[c] = self.format_element(val)

    def new_head(self):
        self._rclass_state = "h"

    def end_head(self):
        self._rclass_state = "b"

    def new_tbody(self):
        self._rclass_state = "b"

    def end_tbody(self):
        self._rclass_state = "b"

    def new_foot(self):
        self._rclass_state = "f"

    def end_foot(self):
        self._rclass_state = "b"

    def new_hcell(self):
        self._rows[-1].inds.append(self._rows[-1].cursor)
        self._rows[-1].row.append("")
        self._rows[-1].cell_type.append("h")

    def end_hcell(self):
        self.end_cell()

    def new_bold(self):
        self._bold_state = True

    def end_bold(self):
        self._bold_state = False



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy
import hashlib


def dr_rms(y):
    n = y.shape
    return numpy.sqrt(2.0 * numpy.sum(y**2.0, 0) / float(n[0]))


def u_rms(y):
    n = y.shape
    return numpy.sqrt(numpy.sum(y**2.0, 0) / float(n[0]))


def decibel_u(y, ref):
    return 20.0 * numpy.log10(y / ref)


def decibel_p(y, ref):
    return 10.0 * numpy.log10(y / ref)


def audio_min():
    return 1.0 / (2.0**24)


def audio_min16():
    return 1.0 / (2.0**16)


def max_dynamic(bit):
    return 20.0 * numpy.log10(2.0**bit)


def normalize(y, ml=1.0):
    m = numpy.max(numpy.abs(y))
    return ml * (y * (1.0 / m))

# compute the sha1 fingerprint of the track
# Do not modify this function !!


def sha1_track_v1(y, ext_code=0):
    n = y.shape
    if n[0] <= 44100 * 2:
        shat = hashlib.sha1(y + ext_code).hexdigest()
    else:
        shat = hashlib.sha1(y[::100, :] + ext_code).hexdigest()

    #print( shat )
    return shat



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import subprocess
import sys
import os
import re
import hashlib
import numpy as np

from dr14meter.audio_decoder import AudioDecoder
from dr14meter.dr14_global import get_ffmpeg_cmd

if sys.version_info[0] == 2:
    import ConfigParser
else:
    import configparser as ConfigParser

from io import StringIO


# Test example !!!!!
# a = subprocess.check_output( [ "ffprobe" , "-show_format" , "/media/esterno_xfs/data/Musica/Musica/aavv/01-blitzkrieg_bop_160_lame_abr.mp3" ] , stderr=subprocess.STDOUT , shell=False )


class UnreadableAudioFileException(Exception):
    pass


class RetrieveMetadata:

    def __init__(self):
        self._album = {}
        self._artist = {}
        self._tracks = {}
        self._disk_nr = []

        if get_ffmpeg_cmd() == "ffmpeg":
            self.__ffprobe_cmd = "ffprobe"
            self.__scan_file = self.scan_file_ffprobe
        elif get_ffmpeg_cmd() == "avconv":
            self.__ffprobe_cmd = "avprobe"
            self.__scan_file = self.scan_file_avprobe

        self.__scan_file = self.scan_file_orig

    def scan_dir(self, dir_name, files_list=None):

        self._album = {}
        self._tracks = {}
        self._artist = {}
        self._disk_nr = []

        if files_list == None:
            dir_name = os.path.abspath(dir_name)
            files_list = sorted(os.listdir(dir_name))

        ad = AudioDecoder()

        for file_name in files_list:

            (fn, ext) = os.path.splitext(file_name)
            full_file = os.path.join(dir_name, file_name)

            if ext in ad.formats:
                try:
                    self.scan_file(full_file)
                except UnreadableAudioFileException as uafe:
                    pass

    def get_scan_file(self):
        return self.__scan_file

    scan_file = property(get_scan_file)

    def match_repetitive_title(self, data_txt):
        re_flags = (re.MULTILINE | re.IGNORECASE | re.UNICODE)
        m = re.search(r"(\S[\S ]+\S)\s*;\s*\1", data_txt, re_flags)
        if m != None:
            return m.group(1)
        else:
            return data_txt

    def scan_file_orig(self, file_name):

        try:
            data_txt = subprocess.check_output([self.__ffprobe_cmd, "-show_format", "-show_streams", file_name],
                                               stderr=subprocess.STDOUT, shell=False)
        except:
            data_txt = "ffprobe ERROR"

        (foo, f_key) = os.path.split(file_name)

        if data_txt != "ffprobe ERROR":
            try:
                data_txt = data_txt.decode(encoding='UTF-8')
            except:
                data_txt = data_txt.decode(encoding='ISO-8859-1')
        else:
            self._tracks[f_key] = None
            raise UnreadableAudioFileException("problematic file: file_name")

        track = {}

        track['file_name'] = file_name

        re_flags = (re.MULTILINE | re.IGNORECASE | re.UNICODE)

        pattern = r"[ \t\f\v]*([\S \t\f\v]+\S).*$"

        m = re.search(r"^\s*track\s*\:\s*(\d+).*$", data_txt, re_flags)
        if m != None:
            #track['nr'] = int( m.group(1) )
            track['track_nr'] = int(m.group(1))

        m = re.search(r"^\s*album\s*\:%s" % pattern, data_txt, re_flags)
        if m != None:
            #print( m.group(1) )
            track['album'] = self.match_repetitive_title(m.group(1))
            self._album.setdefault(track['album'], 0)
            self._album[track['album']] += 1

        m = re.search(r"^\s*title\s*\:%s" % pattern, data_txt, re_flags)
        if m != None:
            track['title'] = self.match_repetitive_title(m.group(1))

        m = re.search(r"^\s*artist\s*\:%s" % pattern, data_txt, re_flags)
        if m != None:
            track['artist'] = self.match_repetitive_title(m.group(1))
            self._artist.setdefault(track['artist'], 0)
            self._artist[track['artist']] += 1

        m = re.search(r"^\s*genre\s*\:%s" % pattern, data_txt, re_flags)
        if m != None:
            track['genre'] = self.match_repetitive_title(m.group(1))

        m = re.search(r"^\s*date\s*\:\s*(\d+).*$", data_txt, re_flags)
        if m != None:
            track['date'] = m.group(1)

        m = re.search(r"^\s*disc\s*:\s*(\d+).*$", data_txt, re_flags)
        if m != None:
            track['disk_nr'] = int(m.group(1))

        m = re.search(r"^size=\s*(\d+)\s*$", data_txt, re_flags)
        if m != None:
            track['size'] = m.group(1)

        m = re.search(r"^bit_rate=\s*(\d+)\s*$", data_txt, re_flags)
        if m != None:
            track['bitrate'] = m.group(1)

        m = re.search(r"^duration=\s*(\d+\.\d+)\s*$", data_txt, re_flags)
        if m != None:
            track['duration'] = float(m.group(1))

        self.__read_stream_info(data_txt, track)

        self._tracks[f_key] = track

    def scan_file_ffprobe(self, file_name):

        try:
            data_txt = subprocess.check_output([self.__ffprobe_cmd, "-show_format", "-show_streams", file_name],
                                               stderr=subprocess.STDOUT, shell=False)
        except:
            data_txt = "ffprobe ERROR"

        if data_txt != "ffprobe ERROR":
            try:
                data_txt = data_txt.decode(encoding='UTF-8')
            except:
                data_txt = data_txt.decode(encoding='ISO-8859-1')

        track = {}
        (foo, f_key) = os.path.split(file_name)

        track['file_name'] = file_name

        re_flags = (re.MULTILINE | re.IGNORECASE | re.UNICODE)

        m = re.search(r"\[FORMAT\](.*)\[/FORMAT\]",
                      data_txt, re_flags | re.DOTALL)
        if m != None:
            format_tags = m.group(1)
        else:
            self._tracks[f_key] = None
            raise UnreadableAudioFileException("problematic file: file_name")

        pattern = r"[ \t\f\v]*([\S \t\f\v]+\S).*$"

        m = re.search(r"^TAG:track=\s*(\d+).*$", format_tags, re_flags)
        if m != None:
            track['track_nr'] = int(m.group(1))

        m = re.search(r"^TAG:disc=\s*(\d+).*$", format_tags, re_flags)
        if m != None:
            track['disk_nr'] = int(m.group(1))
            self._disk_nr.append(int(int(m.group(1))))

        m = re.search(r"^TAG:GENRE=%s" % pattern, format_tags, re_flags)
        if m != None:
            track['genre'] = m.group(1)

        m = re.search(r"^TAG:DATE=\s*(\d+).*$", format_tags, re_flags)
        if m != None:
            track['date'] = m.group(1)

        m = re.search(r"^TAG:ARTIST=%s" % pattern, format_tags, re_flags)
        if m != None:
            self._artist.setdefault(m.group(1), 0)
            self._artist[m.group(1)] += 1
            track['artist'] = m.group(1)

        m = re.search(r"^TAG:TITLE=%s" % pattern, format_tags, re_flags)
        if m != None:
            track['title'] = m.group(1)

        m = re.search(r"^TAG:ALBUM=%s" % pattern, format_tags, re_flags)
        if m != None:
            self._album.setdefault(m.group(1), 0)
            self._album[m.group(1)] += 1
            track['album'] = m.group(1)

        m = re.search(r"^size=\s*(\d+)\s*$", format_tags, re_flags)
        if m != None:
            track['size'] = m.group(1)

        m = re.search(r"^bit_rate=\s*(\d+)\s*$", format_tags, re_flags)
        if m != None:
            track['bitrate'] = m.group(1)

        self.__read_stream_info(data_txt, track)

        self._tracks[f_key] = track

    def scan_file_avprobe(self, file_name):
        try:
            data_txt = subprocess.check_output(
                [self.__ffprobe_cmd, "-show_format", "-show_streams", file_name], stderr=subprocess.STDOUT, shell=False)
        except:
            data_txt = "ffprobe ERROR"

        if data_txt != "ffprobe ERROR":
            try:
                data_txt = data_txt.decode(encoding='UTF-8')
            except:
                data_txt = data_txt.decode(encoding='ISO-8859-1')

        track = {}
        (foo, f_key) = os.path.split(file_name)

        track['file_name'] = file_name

        re_flags = (re.MULTILINE | re.IGNORECASE | re.UNICODE)

        m = re.search(r"(\[format\].*)", data_txt, re_flags | re.DOTALL)
        if m != None:
            format_tags = m.group(1)
        else:
            self._tracks[f_key] = None
            raise UnreadableAudioFileException("problematic file: file_name")

        buf = StringIO(format_tags)
        config = ConfigParser.ConfigParser()
        config.readfp(buf)

        try:
            track['title'] = config.get("format.tags", "title")
            self._album.setdefault(m.group(1), 0)
            self._album[m.group(1)] += 1
        except ConfigParser.NoOptionError:
            pass

        try:
            track['track_nr'] = config.getint("format.tags", "track")
        except ConfigParser.NoOptionError:
            pass

        try:
            track['disk_nr'] = config.get("format.tags", "disc")
            track['disk_nr'] = int(
                re.search(r"(\d+)", track['disk_nr']).group(1))
            self._disk_nr.append(track['disk'])
        except ConfigParser.NoOptionError:
            pass
        except:
            del track['disk_nr']

        try:
            track['genre'] = config.get("format.tags", "genre")
        except ConfigParser.NoOptionError:
            pass

        try:
            track['date'] = config.get("format.tags", "date")
        except ConfigParser.NoOptionError:
            pass

        try:
            track['artist'] = config.get("format.tags", "artist")
            self._artist.setdefault(m.group(1), 0)
            self._artist[m.group(1)] += 1
        except ConfigParser.NoOptionError:
            pass

        try:
            track['album'] = config.get("format.tags", "album")
        except ConfigParser.NoOptionError:
            pass

        try:
            track['size'] = str(int(config.getfloat("format.tags", "size")))
        except ConfigParser.NoOptionError:
            pass

        try:
            track['bitrate'] = str(
                int(config.getfloat("format.tags", "bit_rate")))
        except ConfigParser.NoOptionError:
            pass

        self.__read_stream_info(data_txt, track)

        self._tracks[f_key] = track

    def __read_stream_info(self, data_txt, track):

        re_flags = (re.MULTILINE | re.IGNORECASE | re.UNICODE)

        ##########################################
        # string examples:
        # Audio: flac, 44100 Hz, stereo, s16
        # Stream #0:0(und): Audio: alac (alac / 0x63616C61), 44100 Hz, 2 channels, s16, 634 kb/s
        # Stream #0:0(und): Audio: aac (LC) (mp4a / 0x6134706D), 44100 Hz, stereo, fltp, 255 kb/s (default
        # Stream #0:0: Audio: flac, 44100 Hz, stereo, s16

        m = re.search(r"Stream.*Audio:(.*)$", data_txt, re_flags)
        if m != None:
            fmt = m.group(1)

        fmt = re.split(",", fmt)

        #print( fmt )
        track['codec'] = re.search(r"\s*(\w+)", fmt[0], re_flags).group(1)
        track['sampling_rate'] = re.search(
            r"\s*(\d+)", fmt[1], re_flags).group(1)
        track['channel'] = re.search(
            r"^\s*([\S][\s|\S]*[\S])\s*$", fmt[2], re_flags).group(1)

        m = re.search(r"\((\d+) bit\)", fmt[3], re_flags)
        if m != None:
            track['bit'] = m.group(1)
        else:
            m = re.search(r"(\d+)", fmt[3], re_flags)
            if m != None:
                track['bit'] = m.group(1)
            else:
                track['bit'] = "16"

    def album_len(self):
        return len(self._tracks)

    def get_album_cnt(self):
        return len(self._album)

    def get_disk_nr(self):
        if len(self._disk_nr) > 0:
            return self._disk_nr[0]
        else:
            return None

    def get_album_list(self):
        return self._album

    def get_album_title(self):

        if len(self._album) > 1:
            return "Various"
        elif len(self._album) == 0:
            return None
        else:
            for k in self._album.keys():
                res = k
            return res

    def track_unreadable_failure(self, file_name):
        if self._tracks[file_name] == None:
            return True
        else:
            return False

    def get_album_sha1(self, title=None):

        if title == None:
            p_title = self.get_album_title()
        else:
            p_title = title

        if sys.version_info[0] == 2:
            str_conv = unicode
        else:
            str_conv = str

        key_string = str_conv("")
        #key_string = key_string + str_conv( p_title ) + str_conv( self.get_album_artist() )

        d = np.float64(0.0)
        s = np.float64(0.0)

        for track in sorted(self._tracks.keys()):

            if self._tracks[track] == None:
                continue

            if not self._tracks[track].get('size', False) or not self._tracks[track].get('codec', False):
                continue

            if title != None and not self._tracks[track]["album"] != title:
                continue

            #key_string = key_string + str_conv( track )
            key_string = key_string + str_conv(self._tracks[track]['size'])
            key_string = key_string + str_conv(self._tracks[track]['codec'])
            key_string = key_string + str_conv(self._tracks[track]['duration'])
            key_string = key_string + \
                str_conv(int(self._tracks[track]['bitrate']))
            d += np.float64(self._tracks[track]['duration'])
            s += np.float64(self._tracks[track]['size'])

        key_string = key_string + str_conv(d)
        key_string = key_string + str_conv(s)

        sa = np.frombuffer(bytearray(key_string.encode("utf8")), dtype=np.int8)

        #print( np.sum( sa ) )
        #print( len( sa ) )
        #print( len(key_string) )

        #print( bytearray( key_string.encode("utf8") ) )

        sha1 = hashlib.sha1(sa).hexdigest()
        #print( sha1 )
        return sha1

    def get_album_artist_old(self):

        if len(self._artist) > 1:
            return "Various Artists"
        elif len(self._artist) == 0:
            return None
        else:
            for k in self._artist.keys():
                res = k
            return res

    def get_album_artist(self, album=None):

        if album == None:
            return [self.get_album_artist_old()]

        artists = []
        for track in self._tracks.keys():
            if track["album"] == album:
                if not track["artist"] in artists:
                    artists.append(track["artist"])

        return artists

    def get_value(self, file_name, field):

        f = self._tracks.get(file_name, None)

        if f == None:
            return None

        return f.get(field, None)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from dr14meter.audio_math import *
from dr14meter.plot.my_time_formatter import *
from dr14meter.out_messages import *

try:
    import matplotlib
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
    from matplotlib.widgets import *
    from pylab import *
except:
    pass


def on_select(vmin, vmax):
    dr14_log_debug("on_select: %f , %f " % (vmin, vmax))
    if vmin == vmax:
        return
    plot_track_classic(plot_str=on_select.plot_str,
                       start_time=vmin, end_time=vmax)
    on_select.plot_str.plot()


def wheel_moved(event):
    dr14_log_debug("mouse_pressed: step: %d " % event.step)

    (start_time, end_time) = wheel_moved.plot_str.move(event.step)
    plot_track_classic(plot_str=wheel_moved.plot_str,
                       start_time=start_time, end_time=end_time)
    wheel_moved.plot_str.plot()


def mouse_pressed(event):

    dr14_log_debug("mouse_pressed: ")

    if event.button == 3:
        dr14_log_debug("mouse_pressed: button 3")

        (start_time, end_time) = mouse_pressed.plot_str.zoom_out()
        plot_track_classic(plot_str=mouse_pressed.plot_str,
                           start_time=start_time, end_time=end_time)
        mouse_pressed.plot_str.plot()

        dr14_log_debug("mouse_pressed: start - end:  %f %f " %
                       (start_time, end_time))


class PltTrackStruct:

    def __init__(self, plot_mode='fill', Y=[], Fs=44100, sz=0, ch=0):
        self.ch = ch
        self.tot_time = 0

        self.t = None
        self.tb = None
        self.start_time = 0
        self.end_time = 0

        self.start_block = 0
        self.end_block = 0

        self.mp = None
        self.mn = None

        self.Y = Y
        self.Fs = Fs
        self.plot_mode = plot_mode  # "fill" or "curve"
        self.ax = None
        self.rebuild = False
        self.first_sample = 0
        self.sz_section = 0
        self.lines = []
        self.fig = None

    def start(self):
        self.plot()
        pyplot.show()

    def zoom_out(self):
        start_time = self.start_time
        end_time = self.end_time

        m = start_time + (end_time - start_time) / 2.0
        new_tot_time = 2.0 * (end_time - start_time)

        start_time = m - new_tot_time / 2.0
        end_time = m + new_tot_time / 2.0

        return (start_time, end_time)

    def move(self, sign):

        factor = 0.05

        delta = self.end_time - self.start_time

        start_time = self.start_time + sign * delta * factor
        end_time = self.end_time + sign * delta * factor

        if start_time < 0 or end_time > self.tot_time:
            start_time = self.start_time
            end_time = self.end_time

        return (start_time, end_time)

    def plot(self):

        new_flag = False

        if self.fig == None:
            self.fig = pyplot.figure()
            self.ax = [i for i in range(self.ch)]
            self.lines = [i for i in range(self.ch)]
            self.span = []
            new_flag = True

        for j in range(self.ch):

            if new_flag:
                self.ax[j] = self.fig.add_subplot(210 + j + 1)

            if type(self.lines[j]) is matplotlib.collections.PolyCollection:
                self.lines[j].remove()
            else:
                while (not new_flag) and len(self.lines[j]) > 0:
                    l = self.lines[j].pop(0)
                    l.remove()

            if self.plot_mode == 'fill':
                self.lines[j] = self.ax[j].fill_between(self.tb[self.start_block:self.end_block],
                                                        self.mp[
                                                            self.start_block:self.end_block, j],
                                                        self.mn[self.start_block:self.end_block, j])
            else:
                self.lines[j] = self.ax[j].plot(
                    self.t, self.Y[self.first_sample: self.first_sample + self.sz_section, j], 'b')

            self.ax[j].axis([self.start_time, self.end_time, -1.0, 1.0])

            delta_time = self.end_time - self.start_time

            if delta_time < 1.5:
                milli_sec = True
            else:
                milli_sec = False

            self.ax[j].xaxis.set_major_formatter(
                MyTimeFormatter(milli_sec=milli_sec))

            self.ax[j].grid(True)

            pyplot.title("Channel %d" % (j + 1))
            pyplot.xlabel('Time [min:sec]')
            pyplot.ylabel('Amplitude')

            if new_flag:
                onsel = on_select
                onsel.plot_str = self
                self.span.append(SpanSelector(
                    self.ax[j], onsel, 'horizontal', button=1))

                m_p = mouse_pressed
                m_p.plot_str = self
                connect('button_press_event', m_p)

                w_m = wheel_moved
                w_m.plot_str = self
                connect('scroll_event', w_m)

        self.fig.canvas.draw()


def plot_track_classic(Y=None, Fs=None, plot_str=None, utime=0.02, time_lim=5, start_time=0.0, end_time=-1.0):

    if Y is None and Fs is None and plot_str is None:
        raise

    if plot_str is None and Y is not None and Fs is not None:
        s = Y.shape

        if len(Y.shape) > 1:
            ch = s[1]
        else:
            ch = 1

        plot_str = PltTrackStruct(ch=ch)
        plot_str.Y = Y
        plot_str.Fs = Fs
        plot_str.tot_time = s[0] * 1.0 / Fs

    elif plot_str != None:
        Fs = plot_str.Fs
        Y = plot_str.Y

        s = Y.shape
        if len(Y.shape) > 1:
            ch = s[1]
        else:
            ch = 1
    else:
        raise

    sz_section = s[0]
    first_sample = 0

    if end_time > s[0] * 1.0 / Fs:
        end_time = s[0] * 1.0 / Fs

    if start_time < 0:
        start_time = 0.0

    if start_time >= 0.0 and end_time > start_time:
        sz_section = int((end_time - start_time) * Fs)
        first_sample = int(start_time * Fs)
        plot_str.start_time = start_time
        plot_str.end_time = end_time
    else:
        plot_str.start_time = 0.0
        plot_str.end_time = sz_section * 1.0 / Fs

    block_len = utime

    samples_block = int(block_len * Fs)
    sz = int(sz_section / samples_block)

    curr_sample = first_sample

    if plot_str.mp is not None and plot_str.mn is not None and sz_section >= time_lim * Fs:
        plot_str.plot_mode = "fill"
        plot_str.start_block = int(plot_str.start_time / block_len)
        plot_str.end_block = int(plot_str.end_time / block_len)

        return plot_str

    if sz_section > time_lim * Fs:
        plot_str.plot_mode = "fill"
        plot_str.mp = zeros((sz, ch))
        plot_str.mn = zeros((sz, ch))
        plot_str.tb = start_time + np.arange(0, sz) * block_len

        for i in range(sz):
            plot_str.mp[i, :] = np.max(
                Y[curr_sample:curr_sample + samples_block, :], 0)
            plot_str.mn[i, :] = np.min(
                Y[curr_sample:curr_sample + samples_block, :], 0)
            curr_sample = curr_sample + samples_block

        plot_str.start_block = 0
        plot_str.end_block = sz

    else:
        plot_str.plot_mode = "curve"
        plot_str.t = start_time + np.arange(0, sz_section) * 1.0 / Fs
        plot_str.first_sample = first_sample
        plot_str.sz_section = sz_section

    return plot_str



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
import dr14meter.audio_math as am
import dr14meter.out_messages as om
from dr14meter.plot.my_time_formatter import MyTimeFormatter

import math
import time

try:
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
except:
    pass


def dynamic_vivacity(Y, Fs, Plot=True):

    time_a = time.time()

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    block_size = 1.0

    samples_per_block = int(Fs * block_size)

    seg_cnt = int(math.floor(s[0] / samples_per_block) + 1)

    seg_dyn = numpy.zeros((seg_cnt, ch))

    curr_sample = 0
    for i in range(seg_cnt - 1):
        #r = numpy.arange( curr_sample , curr_sample + samples_per_block )
        rms = am.u_rms(Y[curr_sample: curr_sample + samples_per_block, :])
        mx = numpy.max(Y[curr_sample: curr_sample + samples_per_block, :], 0)
        seg_dyn[i, :] = am.decibel_u(mx, rms)
        iz = (rms < am.audio_min())
        seg_dyn[i, iz] = 0.0
        curr_sample = curr_sample + samples_per_block

    i = seg_cnt - 1
    r = numpy.arange(curr_sample, s[0])

    if r.shape[0] > 0:
        rms = am.u_rms(Y[r, :])
        mx = numpy.max(Y[r, :])
        seg_dyn[i, :] = am.decibel_u(mx, rms)
        iz = (rms < am.audio_min())
        seg_dyn[i, iz] = 0.0

    t = numpy.arange(0.0, math.floor((1.0 / Fs) * s[0]) + 1, step=block_size)

    max_db = numpy.max(seg_dyn)

    non_zeros_ = seg_dyn >= am.audio_min()

    non_zeros = numpy.ones((seg_dyn.shape[0])) == 1

    for i in range(ch):
        non_zeros = numpy.logical_and(non_zeros, non_zeros_[:, i])

    mean = numpy.mean(seg_dyn[non_zeros, :], 0)
    std = numpy.std(seg_dyn[non_zeros, :], 0)

    tot_t = s[0] * 1.0 / Fs

    time_b = time.time()
    om.dr14_log_info("dynamic_vivacity: Clock: %2.8f" % (time_b - time_a))

    if Plot:
        for j in range(ch):

            ax = pyplot.subplot(210 + j + 1)
            pyplot.plot(t.T, seg_dyn[:, j], linewidth=2, color="b")
            pyplot.grid(True)

            time_x = numpy.array([0, tot_t])
            mean_y = numpy.array([mean[j], mean[j]])

            pyplot.plot(time_x, mean_y, linewidth=2, color="g")

            ax.xaxis.set_major_formatter(MyTimeFormatter())
            pyplot.xticks(rotation=25)

            std_a_y = numpy.array([mean[j], mean[j]] + std[j])
            std_b_y = numpy.array([mean[j], mean[j]] - std[j])

            pyplot.plot(time_x, std_a_y, linewidth=2, ls='--', color='c')
            pyplot.plot(time_x, std_b_y, linewidth=2, ls='--', color='c')

            pyplot.axis([0, tot_t, 0, max_db * 1.15])

            text_rel_pos = 0.2
            pyplot.text(tot_t * 0.1, max_db * text_rel_pos,
                        "mean:     %.3f dB" % mean[j], fontsize=12)
            pyplot.text(tot_t * 0.1, max_db * (text_rel_pos - 0.07),
                        "std dev:  %.3f dB" % std[j], fontsize=12)

            pyplot.title("Channel %d" % (j + 1))
            pyplot.xlabel('Time [min:sec]')
            pyplot.ylabel('Dynamic. [dB]')

        pyplot.show()

    return (mean, std, seg_dyn)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy
import math
from dr14meter.audio_math import *

try:
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
except:
    pass


def compute_hist(Y, Fs, duration=None, bins=100, block_duration=0.2, plot=True, title=None):
    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    saples_per_block = int(Fs * block_duration)

    seg_cnt = int(math.floor(float(s[0]) / saples_per_block)) + 1

    if seg_cnt < 3:
        return False

    curr_sam = 0
    rms = numpy.zeros((seg_cnt, ch))

    #peaks = numpy.zeros((seg_cnt,ch))

    for i in range(seg_cnt - 1):
        #r = numpy.arange( curr_sam , curr_sam + saples_per_block )
        rms[i, :] = u_rms(Y[curr_sam: curr_sam + saples_per_block, :])
        curr_sam = curr_sam + saples_per_block

    i = seg_cnt - 1
    r = numpy.arange(curr_sam, s[0])

    if r.shape[0] > 0:
        rms[i, :] = dr_rms(Y[r, :])

    rms = numpy.sum(rms, 1) / float(ch)

    rms[rms == 0.0] = audio_min16()
    rms = decibel_u(rms, 1.0)

    rms_mean = numpy.mean(rms)
    rms_std = numpy.std(rms)
    rms_max = numpy.max(rms)

    abs_peak = decibel_u(numpy.max(numpy.abs(Y)), 1.0)

    if plot == True:
        (hist, bin_edges, patches) = pyplot.hist(rms, 100, density=True)

        mean_x = numpy.array([rms_mean, rms_mean])
        mean_y = numpy.array([0.0, numpy.max(hist) * 1.01])

        std_x = numpy.array([rms_mean - rms_std, rms_mean + rms_std])
        std_y = numpy.array([numpy.max(hist) * 0.7, numpy.max(hist) * 0.7])

        pyplot.plot(mean_x, mean_y, linewidth=2, color='g')
        pyplot.plot(std_x, std_y, linewidth=2, ls='--', color='c')
        #print( hist )

        pyplot.axis([-92, 0, 0, numpy.max(hist) * 1.05])

        text_rel_pos = 0.9
        pyplot.text(-85, numpy.max(hist) * text_rel_pos,
                    "mean:    %.3f dB" % rms_mean, fontsize=15,)
        pyplot.text(-85, numpy.max(hist) * (text_rel_pos - 0.05),
                    "std dev:  %.3f dB" % rms_std, fontsize=15,)
        pyplot.text(-85, numpy.max(hist) * (text_rel_pos - 0.10),
                    "peak:     %.3f dB" % abs_peak, fontsize=15,)
        pyplot.text(-85, numpy.max(hist) * (text_rel_pos - 0.15),
                    "max rms:  %.3f dB" % rms_max, fontsize=15,)

        pyplot.xlabel('RMS dB')
        pyplot.ylabel('Relative frequency')

        if title != None:
            hist_title = title
        else:
            hist_title = "Hystogram of dynamic"

        pyplot.title(r'%s' % hist_title)

        pyplot.plot(mean_x, mean_y)

        pyplot.grid(True)
        pyplot.show()

    else:
        (hist, bin_edges) = numpy.histogram(rms, bins=bins, density=True)

    return (hist, bin_edges)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011 - 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile

# from dr14meter.compute_dr import *
from dr14meter.audio_track import AudioTrack
# from dr14meter.table import *
from dr14meter.compressor import DynCompressor
from dr14meter.wav_write import wav_write
from dr14meter.read_metadata import RetrieveMetadata
from dr14meter.duration import StructDuration

from dr14meter.plot.dr_histogram import compute_hist
from dr14meter.plot.lev_histogram import compute_lev_hist
from dr14meter.plot.spectrogram import spectrogram
from dr14meter.plot.dynamic_vivacity import dynamic_vivacity
from dr14meter.plot.plot_track import plot_track
from dr14meter.plot.plot_track_classic import plot_track_classic

from dr14meter.out_messages import print_msg


class AudioAnalysis:

    def compute_track(self, file_name):
        self.at = AudioTrack()

        (head,  file_n) = os.path.split(file_name)

        if not self.at.open(file_name):
            return False

        self.file_name = file_name

        self.meta_data = RetrieveMetadata()
        self.meta_data.scan_file(self.file_name)

        self.duration = StructDuration()
        self.duration.set_samples(self.at.Y.shape[0], self.at.Fs)

        self.virt_compute()

        return True

    def getDuration(self):
        return self.duration

    def getMetaData(self):
        return self.meta_data

    def getAudioTrack(self):
        return self.at

    def getFileName(self):
        return self.file_name

    def virt_compute(self):
        raise


class AudioDynVivacity(AudioAnalysis):

    def virt_compute(self):
        (foo, fn) = os.path.split(self.getFileName())

        title = self.getMetaData().get_value(fn, "title")

        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()
        dynamic_vivacity(at.Y, at.Fs)


class AudioDrHistogram(AudioAnalysis):

    def virt_compute(self):

        (foo, fn) = os.path.split(self.getFileName())

        title = self.getMetaData().get_value(fn, "title")

        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()
        compute_hist(at.Y, at.Fs, self.getDuration(), title=title)


class AudioLevelHistogram(AudioAnalysis):

    def virt_compute(self):

        (foo, fn) = os.path.split(self.getFileName())

        title = self.getMetaData().get_value(fn, "title")

        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()
        compute_lev_hist(at.Y, at.Fs, self.getDuration(), title=title)


class AudioSpectrogram(AudioAnalysis):

    def virt_compute(self):

        (foo, fn) = os.path.split(self.getFileName())

        title = self.getMetaData().get_value(fn, "title")

        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()

        spectrogram(at.Y, at.Fs)


class AudioPlotTrack(AudioAnalysis):

    def virt_compute(self):

        (foo, fn) = os.path.split(self.getFileName())
        title = self.getMetaData().get_value(fn, "title")
        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()
        plot_str = plot_track_classic(at.Y, at.Fs)
        plot_str.start()


class AudioPlotTrackDistribution(AudioAnalysis):

    def virt_compute(self):

        (foo, fn) = os.path.split(self.getFileName())
        title = self.getMetaData().get_value(fn, "title")
        print_msg("Track Title: %s " % title)

        at = self.getAudioTrack()
        plot_track(at.Y, at.Fs)


class AudioCompressor(AudioAnalysis):

    def setCompressionModality(self, compression_modality):
        self.compression_modality = compression_modality

    def virt_compute(self):
        (head, file_n) = os.path.split(self.getFileName())

        comp = DynCompressor()
        comp.set_compression_modality(self.compression_modality)

        full_file = os.path.join(
            tempfile.gettempdir(), "%s%s.wav" % (file_n, "-compressed-"))

        at = self.getAudioTrack()
        cY = comp.dyn_compressor(at.Y, at.Fs)

        wav_write(full_file, at.Fs, cY)
        print_msg(
            "The resulting compressed audiotrack has been written in: %s " % full_file)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from dr14meter.audio_math import *
from dr14meter.out_messages import *

import numpy as np

import math
import time


def compute_dr14(Y, Fs, duration=None, Dr_lr=None):

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    if Fs == 44100:
        delta_fs = 60
    else:
        delta_fs = 0

    dr14_log_debug("compute_dr14: Y: Fs: %d ; ch: %d ; shape: %d " %
                   (Fs, ch, s[0]))
    time_a = time.time()

    block_time = 3
    cut_best_bins = 0.2
    block_samples = block_time * (Fs + delta_fs)

    seg_cnt = int(math.floor(s[0] / block_samples) + 1)

    if seg_cnt < 1:
        dr14_log_debug("compute_dr14: EXIT - too short")
        return (0, -100, -100)

    curr_sam = 0
    rms = np.zeros((seg_cnt, ch))
    peaks = np.zeros((seg_cnt, ch))

    for i in range(seg_cnt - 1):
        rms[i, :] = np.sqrt(2.0 * np.sum(Y[curr_sam:curr_sam + block_samples, :]**2.0, 0) /
                            float(block_samples))
        peaks[i, :] = np.max(
            np.abs(Y[curr_sam:curr_sam + block_samples, :]), 0)
        curr_sam = curr_sam + block_samples

    i = seg_cnt - 1

    if curr_sam < s[0]:
        rms[i, :] = dr_rms(Y[curr_sam:s[0] - 1, :])
        peaks[i, :] = np.max(np.abs(Y[curr_sam:s[0] - 1, :]), 0)

    peaks = np.sort(peaks, 0)
    rms = np.sort(rms, 0)

    n_blk = int(math.floor(seg_cnt * cut_best_bins))
    if n_blk == 0:
        n_blk = 1

    r = np.arange(seg_cnt - n_blk, seg_cnt)

    rms_sum = np.sum(rms[r, :]**2, 0)

    ch_dr14 = -20.0 * np.log10(np.sqrt(rms_sum / n_blk)
                               * 1.0 / peaks[seg_cnt - 2, :])

    err_i = np.logical_or(rms_sum < audio_min(),
                          np.abs(ch_dr14) > max_dynamic(24))
    ch_dr14[err_i] = 0.0

    dr14 = round(np.mean(ch_dr14))

    db_peak = decibel_u(np.max(peaks), 1.0)

    #y_rms = np.sqrt(np.sum(((np.sum(Y, 1) / 2)**2)) / Y.shape[0])

    y_rms = dr_rms(Y)
    y_rms = np.mean(y_rms)

    db_rms = decibel_u(y_rms, 1.0)

    if duration != None:
        duration.set_samples(s[0], Fs)

    if Dr_lr != None:
        Dr_lr = ch_dr14

    time_b = time.time()
    dr14_log_info("compute_dr14: Clock: %2.8f" % (time_b - time_a))

    return (dr14, db_peak, db_rms)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011 - 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import codecs
import multiprocessing as mp

from dr14meter.compute_dr14 import compute_dr14
from dr14meter.compute_dr import *
from dr14meter.audio_track import *
from dr14meter.table import *
from dr14meter.read_metadata import RetrieveMetadata
from dr14meter.audio_decoder import AudioDecoder
from dr14meter.duration import StructDuration
from dr14meter.write_dr import WriteDr, WriteDrExtended
from dr14meter.audio_math import sha1_track_v1
from dr14meter.dr14_config import get_collection_dir
from dr14meter.out_messages import print_msg, flush_msg

import dr14meter.dr14_global as dr14

from dr14meter.out_messages import print_msg, print_out


class SharedDrResObj:

    def __init__(self):
        self.track_nr = -1
        self.file_name = ""
        self.dr14 = dr14.min_dr()
        self.dB_peak = 0.0
        self.dB_rms = 0.0
        self.duration = ""
        self.sha1 = ""
        self.fail = False
        self.dir_name = ""


class DynamicRangeMeter:

    def __init__(self):
        self.res_list = []
        self.dir_name = ''
        self.dr14 = 0
        self.meta_data = RetrieveMetadata()
        self.compute_dr = ComputeDR14()
        self.__write_to_local_db = False
        self.coll_dir = os.path.realpath(get_collection_dir())

    def write_to_local_db(self, f=False):
        self.__write_to_local_db = f

    def scan_file(self, file_name):

        at = AudioTrack()

        duration = StructDuration()

        if at.open(file_name):
            self.__compute_and_append(at, file_name)
            return 1
        else:
            return 0

    def scan_dir(self, dir_name):

        if not os.path.isdir(dir_name):
            return 0

        dir_list = sorted(os.listdir(dir_name))

        self.dir_name = dir_name
        self.dr14 = 0

        duration = StructDuration()

        at = AudioTrack()
        for file_name in dir_list:
            full_file = os.path.join(dir_name, file_name)

            #print_msg( full_file )
            if at.open(full_file):
                self.__compute_and_append(at, file_name)

        self.meta_data.scan_dir(dir_name)
        if len(self.res_list) > 0:
            self.dr14 = int(round(self.dr14 / len(self.res_list)))
            return len(self.res_list)
        else:
            return 0

    def __compute_and_append(self, at, file_name):

        duration = StructDuration()

        #( dr14, dB_peak, dB_rms ) = self.compute_dr.compute( at.Y , at.Fs )
        (dr14, dB_peak, dB_rms) = compute_dr14(at.Y, at.Fs, duration)
        sha1 = sha1_track_v1(at.Y, at.get_file_ext_code())

        self.dr14 = self.dr14 + dr14

        res = {'file_name': file_name,
               'dr14': dr14,
               'dB_peak': dB_peak,
               'dB_rms': dB_rms,
               'duration': duration.to_str(),
               'sha1': sha1}

        self.res_list.append(res)

        print_msg(file_name + ": \t DR " + str(int(dr14)))

    def write_to_local_database(self):

        wr = WriteDr()

        if self.__write_to_local_db and os.path.realpath(self.dir_name).startswith(self.coll_dir):
            wr.write_to_local_dr_database(self)

    def fwrite_dr(self, file_name, tm, ext_table=False, std_out=False, append=False, dr_database=True):

        if ext_table:
            wr = WriteDrExtended()
        else:
            wr = WriteDr()

        wr.set_loudness_war_db_compatible(dr_database)

        self.table_txt = wr.write_dr(self, tm)

        if std_out:
            print_out(self.table_txt)
            return

        if append:
            file_mode = "a"
        else:
            file_mode = "w"

        try:
            out_file = codecs.open(file_name, file_mode, "utf-8-sig")
        except:
            print_msg("File opening error [%s] :" %
                      file_name, sys.exc_info()[0])
            return False

        out_file.write(self.table_txt)
        out_file.close()
        return True

    def scan_mp(self, dir_name="", thread_cnt=2, files_list=[]):

        if sys.version_info[0] == 2:
            dir_name = dir_name.decode('utf-8')

        self.dr14 = 0

        job_queue_sh = mp.JoinableQueue(2000)
        res_queue_sh = mp.Queue(2000)

        if files_list == []:
            if not os.path.isdir(dir_name):
                return 0
            dir_list = sorted(os.listdir(dir_name))
            self.dir_name = dir_name
            files_list = None
        else:
            dir_list = sorted(files_list)

        ad = AudioDecoder()

        for file_name in dir_list:
            (fn, ext) = os.path.splitext(file_name)
            if ext in ad.formats:
                job = SharedDrResObj()
                job.file_name = file_name
                job.dir_name = dir_name
                job_queue_sh.put(job)

        threads = [1 for i in range(thread_cnt)]

        job_free = mp.Value('i', 0)

        for t in range(thread_cnt):
            threads[t] = mp.Process(
                target=self.run_mp, args=(job_queue_sh, res_queue_sh))

        for t in range(thread_cnt):
            threads[t].start()

        job_queue_sh.join()

        succ = 0

        self.res_list = []

        #i = 0

        dur = StructDuration()

        while not res_queue_sh.empty():
            res = res_queue_sh.get()
            if res.fail:
                continue
            self.res_list.append({'file_name':   res.file_name,
                                  'dr14':        res.dr14,
                                  'dB_peak':     res.dB_peak,
                                  'dB_rms':      res.dB_rms,
                                  'duration':    dur.float_to_str(res.duration),
                                  'sha1':        res.sha1})

        self.res_list = sorted(self.res_list, key=lambda res: res['file_name'])

        #    i = i + 1

        for d in self.res_list:
            if d['dr14'] > dr14.min_dr():
                self.dr14 = self.dr14 + d['dr14']
                succ = succ + 1

        self.meta_data.scan_dir(dir_name, files_list)

        if len(self.res_list) > 0 and succ > 0:
            self.dr14 = int(round(self.dr14 / succ))
            return succ
        else:
            return 0

    def run_mp(self, job_queue_sh, res_queue_sh):

        at = AudioTrack()
        duration = StructDuration()

        #print_msg("start .... ")

        while True:

            if job_queue_sh.empty():
                return

            job = job_queue_sh.get()

            full_file = os.path.join(job.dir_name, job.file_name)
            #print ( full_file )

            if at.open(full_file):
                (dr14, dB_peak, dB_rms) = compute_dr14(at.Y, at.Fs, duration)
                sha1 = sha1_track_v1(at.Y, at.get_file_ext_code())

                print_msg(job.file_name + ": \t DR " + str(int(dr14)))
                flush_msg()

                job.dr14 = dr14
                job.dB_peak = dB_peak
                job.dB_rms = dB_rms
                job.duration = duration.to_float()
                job.sha1 = sha1

            else:
                job.fail = True
                print_msg("- fail - " + full_file)

            res_queue_sh.put(job)
            job_queue_sh.task_done()



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os.path
import os
import sys

if sys.version_info[0] == 2:
    import ConfigParser
else:
    import configparser as ConfigParser


def get_config_directory(create=True):

    p = os.environ.get('XDG_CONFIG_HOME')

    if p is None or not os.path.isabs(p):
        p = os.path.expanduser('~/.config')

    cfg_dir = os.path.join(p, 'dr14meter')

    if not os.path.isdir(cfg_dir) and create:
        os.mkdir(cfg_dir)

    return cfg_dir


def get_config_file(create=True):
    cfg_dir = get_config_directory()
    cfg_file = "%s/%s" % (cfg_dir, "dr14.cfg")

    if not os.path.isfile(cfg_file) and create:
        write_default_cfg(cfg_file)

    return cfg_file


def write_default_cfg(cfg_file):
    config = ConfigParser.ConfigParser()

    config.add_section('config_version')
    config.set('config_version', 'number', '1')

    config.add_section('database')

    config.set('database', 'enabled', 'False')
    config.set('database', 'path', get_config_directory() + "/dr14.db")
    config.set('database', 'collection_dir', '/')

    with open(cfg_file, 'w') as configfile:
        config.write(configfile)


def set_db_path(full_file_path):
    set_config_field('database', 'path', full_file_path)


def enable_db(flag=True):
    set_config_field('database', 'enabled', str(flag))


def database_exists():
    dbp = get_db_path()
    if os.path.isfile(dbp):
        return True
    else:
        return False


def set_collection_dir(path):
    set_config_field('database', 'collection_dir', path)


def set_config_field(section, field, value):
    cfg_file = get_config_file()
    config = ConfigParser.ConfigParser()
    config.read(cfg_file)
    config.set(section, field, value)
    with open(cfg_file, 'w') as configfile:
        config.write(configfile)


def get_db_path():
    return get_config_filed('database', 'path')


def db_is_enabled():
    s = get_config_filed('database', 'enabled')
    return s.lower() in ["yes", "true", "t", "1", "ok", "yup"]


def get_collection_dir():
    return get_config_filed('database', 'collection_dir')


def get_config_filed(section, field):
    config = ConfigParser.ConfigParser()
    cfg_file = get_config_file()
    config.read(cfg_file)
    return config.get(section, field)



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib.metadata
import threading
import subprocess
import sys

from dr14meter.out_messages import print_msg


lock_ver = threading.Lock()

ffmpeg_cmd = None


def dr14_version():
    try:
        return importlib.metadata.version('dr14meter')
    except ModuleNotFoundError:
        print_msg("The %s function require the installation of MatPlotLib")
    return 'UNDEF'


def min_dr():
    return -10000


def get_exe_name():
    return "dr14meter"


def get_name_version():
    return f'dr14meter {dr14_version()}'


def get_ffmpeg_cmd():

    global ffmpeg_cmd

    if ffmpeg_cmd is None:
        for cmd in ['ffmpeg', 'avconv']:
            if sys.platform.startswith('win'):
                cmd_exists = subprocess.call(f'where {cmd}', shell=True, stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE) == 0
            else:
                cmd_exists = subprocess.call(f'type {cmd}', shell=True, stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE) == 0
            if cmd_exists:
                ffmpeg_cmd = cmd
                break
    # issue #4: if nothing works fallback to ffmpeg... maybe it will work for some reason. Still better than None
    if ffmpeg_cmd is None:
        ffmpeg_cmd = 'ffmpeg'
    return ffmpeg_cmd


def get_home_url():
    return "https://github.com/pe7ro/dr14meter"


def test_lib(lib_name, fun_name):
    try:
        # __import__(lib_name)
        importlib.import_module(lib_name)
    except ImportError:
        print_msg(f"The {fun_name} function require the installation of {lib_name}")
        return False

    return True


def test_matplotlib_modules(fun_name):
    return test_lib('matplotlib', fun_name)


def test_mutagen(fun_name):
    return test_lib('mutagen', fun_name)


def test_hist_modules():
    return test_lib('matplotlib', 'histogram')


def test_compress_modules():
    return test_lib('scipy', 'compression')




# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# __all__ = ["main"]

import sys

# from src import dr14meter

if (__package__ is None or len(__package__) == 0) and not getattr(sys, 'frozen', False):
    import pathlib
    path = pathlib.Path(__file__).resolve()
    sys.path.insert(0, str(path.parent.parent))

import dr14meter.dr14_tmeter as dr14
# import .dr14meter


def main():
    """Entry point for the application script"""
    dr14.main()


if __name__ == '__main__':
    main()





# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys

from dr14meter.database.database import dr_database_singletone


def my_dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class query:

    def __init__(self):
        self.keys = []
        self.__params = list([30])

    def append_parameter(self, p):
        self.__params.append(p)

    def set_parameter(self, i, p):
        self.__params[i] = p

    def get_parameter(self, i):
        return self.__params[i]

    def get_limit(self):
        return self.__params[0]

    def set_limit(self, limit):
        self.__params[0] = limit

    limit = property(get_limit, set_limit)

    def get_query(self):
        NotImplementedError(
            "%s : is virtual and must be overridden." % sys._getframe().f_code.co_name)

    def exec_query(self):
        db = dr_database_singletone().get()
        return db.query(self.get_query(), (self.limit,), dict_factory_arg=my_dict_factory)

    def get_col_keys(self):
        return self.keys


class query_top_dr(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["DR", "Title", "Artist", "Album"]

    def get_query(self):
        q = """
        select track.id as id , track.title as Title , dr.dr as DR , Album.title as Album , Artist.name as Artist 
           from track inner join DR_Track on DR_Track.idtrack = track.id 
                inner join dr on dr.id = DR_Track.iddr
                inner join Album_Track on Album_Track.idtrack = track.id
                inner join Album on Album.id = Album_Track.idalbum
                inner join Artist_Track on Artist_Track.idtrack = track.id
                inner join Artist on Artist.id = Artist_Track.IdArtist
                order by dr desc
                limit ? ;
        """

        return q


class query_top_albums_dr(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["DR", "Album_Title", "Artist"]

    def get_query(self):
        q = """
        select album.title as Album_Title , dr.dr as DR , album.id as id , Artist.name as Artist
           from album inner join DR_Album on DR_Album.idalbum = album.id 
                inner join dr on dr.id = DR_Album.iddr   
                inner join Artist_Album on album.id = Artist_Album.IdAlbum
                inner join Artist on Artist.Id = Artist_Album.IdArtist       
                order by dr desc
                limit ? ;
        """

        return q


class query_worst_albums_dr(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["DR", "Album_Title", "Artist"]

    def get_query(self):
        q = """
        select album.title as Album_Title , dr.dr as DR , album.id as id , Artist.name as Artist
           from album inner join DR_Album on DR_Album.idalbum = album.id
                inner join dr on dr.id = DR_Album.iddr
                inner join Artist_Album on album.id = Artist_Album.IdAlbum
                inner join Artist on Artist.Id = Artist_Album.IdArtist
                order by dr asc
                limit ? ;
        """

        return q


class query_worst_dr(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["DR", "Title", "Artist", "Album"]

    def get_query(self):
        q = """
        select track.id as id , track.title as Title , dr.dr as DR , Album.title as Album , Artist.name as Artist 
           from track inner join DR_Track on DR_Track.idtrack = track.id 
                inner join dr on dr.id = DR_Track.iddr
                inner join Album_Track on Album_Track.idtrack = track.id
                inner join Album on Album.id = Album_Track.idalbum
                inner join Artist_Track on Artist_Track.idtrack = track.id
                inner join Artist on Artist.id = Artist_Track.IdArtist
                order by dr asc
                limit ? ;
        """

        return q


class query_top_artists(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["Mean_DR", "Artist", "Track_Count"]
        self.append_parameter(10)

    def set_min_track(self, mt):
        self.set_parameter(1, mt)

    def get_min_track(self):
        return self.get_parameter(1)

    min_track = property(get_min_track, set_min_track)

    def exec_query(self):
        db = dr_database_singletone().get()
        return db.query(self.get_query(), (self.min_track, self.limit), dict_factory_arg=my_dict_factory)

    def get_query(self):
        q = """
        select Artist, Mean_DR , Track_Count from 
        ( 
            select artist.name as Artist , avg( dr.dr ) as Mean_DR , count( track.id ) as Track_Count
               from track inner join DR_Track on DR_Track.idtrack = track.id 
                      inner join dr on dr.id = DR_Track.iddr 
                      inner join Artist_Track on Artist_Track.idtrack = track.id
                      inner join Artist on Artist.id = Artist_Track.IdArtist
                      group by artist                  
        )
        where Track_Count >= ? 
        order by mean_dr desc  
        limit ? ;
        """

        return q


class query_dr_histogram(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["DR", "Freq"]

    def exec_query(self):
        db = dr_database_singletone().get()
        return db.query(self.get_query(), (), dict_factory_arg=my_dict_factory)

    def get_query(self):
        q = """
        select dr.dr as DR , count(dr.dr) as Freq 
            from  DR_Track inner join dr on dr.id = DR_Track.iddr
            where dr.dr >= 0 
            group by (dr.dr) 
            order by (dr) ;
        """

        return q


class query_date_dr_evolution(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["Date", "Mean"]

    def exec_query(self):
        db = dr_database_singletone().get()
        return db.query(self.get_query(), (), dict_factory_arg=my_dict_factory)

    def get_query(self):
        q = """
        select date.date as Date , avg( dr.dr ) as Mean
           from track inner join DR_Track on DR_Track.idtrack = track.id 
                  inner join dr on dr.id = DR_Track.iddr
                  inner join Date_Track on Date_Track.idtrack = track.id
                  inner join date on date.id = Date_Track.iddate
                  where dr.dr >= 0  
                  group by date 
                  order by date ;
        """

        return q


class query_dr_codec(query):

    def __init__(self):
        query.__init__(self)
        self.keys = ["Codec", "Mean_DR", "Codec_Freq"]

    def exec_query(self):
        db = dr_database_singletone().get()
        return db.query(self.get_query(), (), dict_factory_arg=my_dict_factory)

    def get_query(self):
        q = """
        select codec.name as Codec , avg( dr.dr ) as Mean_DR , count( Codec_Track.IdCodec ) as Codec_Freq 
           from track inner join DR_Track on DR_Track.idtrack = track.id 
                  inner join dr on dr.id = DR_Track.iddr 
                  inner join Codec_Track on Codec_Track.idtrack = track.id
                  inner join Codec on Codec.id = Codec_Track.IdCodec 
                  group by name 
                  order by mean_dr desc ;  
        """

        return q



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import matplotlib
    import matplotlib.pyplot as pyplot
    import matplotlib.mlab as mlab
except:
    pass


try:
    class MyTimeFormatter(matplotlib.ticker.Formatter):

        def __init__(self, utime=1.0, milli_sec=False):
            self.utime = utime
            self.milli_sec = milli_sec

        def __call__(self, x, pos=None):
            minu = int(self.utime * x / 60)
            sec = int(self.utime * x - minu * 60)
            msec = int(1000 * (self.utime * x - int(self.utime * x)))

            if self.milli_sec:
                return "%02d:%02d.%03d" % (minu, sec, msec)
            else:
                return "%02d:%02d" % (minu, sec)
except:
    class MyTimeFormatter:

        def __init__(self):
            raise



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011 - 2012  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import dr14meter.dr14_global as dr14
import dr14meter.table as table

from dr14meter.database.database import dr_database_singletone


class WriteDr:

    def __init__(self):
        self.__dr_database_compatible = True

    def set_loudness_war_db_compatible(self, f):
        self.__dr_database_compatible = f

    def get_loudness_war_db_compatible(self):
        return self.__dr_database_compatible

    def write_to_local_dr_database(self, drm):
        db = dr_database_singletone().get()
        db.open_insert_session()

        album_title = drm.meta_data.get_album_title()

        if album_title == None:
            (head, album_title) = os.path.split(drm.dir_name)

        album_sha1 = drm.meta_data.get_album_sha1()
        album_artist = drm.meta_data.get_album_artist()

        db.insert_album(album_sha1, album_title,
                        int(drm.dr14),
                        disk_nr=drm.meta_data.get_disk_nr(), artist=album_artist[0])

        for i in range(len(drm.res_list)):

            curr_file_name = drm.res_list[i]['file_name']

            if drm.meta_data.track_unreadable_failure(curr_file_name):
                continue

            track_sha1 = drm.res_list[i]['sha1']
            title = drm.meta_data.get_value(curr_file_name, 'title')
            dr = drm.res_list[i]['dr14']
            rms = drm.res_list[i]['dB_rms']
            peak = drm.res_list[i]['dB_peak']
            duration = drm.meta_data.get_value(curr_file_name, 'duration')
            size = drm.meta_data.get_value(curr_file_name, 'size')
            bit = drm.meta_data.get_value(curr_file_name, 'bit')
            bitrate = drm.meta_data.get_value(curr_file_name, 'bitrate')
            sampling_rate = drm.meta_data.get_value(
                curr_file_name, 'sampling_rate')
            codec = drm.meta_data.get_value(curr_file_name, 'codec')
            artist = drm.meta_data.get_value(curr_file_name, 'artist')
            genre = drm.meta_data.get_value(curr_file_name, 'genre')
            date = drm.meta_data.get_value(curr_file_name, 'date')
            track_nr = drm.meta_data.get_value(curr_file_name, 'track_nr')

            if title == None:
                title = curr_file_name

            db.insert_track(track_sha1, title,
                            dr, rms, peak, duration,
                            codec,  bit, bitrate, sampling_rate,
                            album_sha1, artist,
                            genre, date, track_nr, size)

        db.commit_insert_session()

    def write_query_result(self, res_dl, tm, table_title, desired_keys=None, desired_keys_titles=None):

        if len(res_dl) == 0:
            return ""

        if desired_keys is not None:
            keys = desired_keys
        else:
            keys = res_dl[0].keys()

        if len(keys) == 0:
            return ""

        tm.new_table()
        tm.col_cnt = len(keys)

        tm.append_separator_line()

        tm.add_title(table_title)

        tm.new_tbody()

        tm.append_separator_line()
        tm.append_row(keys, 'h')
        tm.append_separator_line()

        for row in res_dl:
            tm.append_row([row[k] for k in keys])

        tm.end_tbody()

        tm.end_table()

        return tm.write_table()

    def write_dr(self, drm, tm):

        (head, album_dir) = os.path.split(drm.dir_name)

        tm.new_table()

        tm.new_head()

        tm.append_separator_line()
        tm.add_title(" Analyzed folder:  " + album_dir)

        tm.end_head()

        tm.new_tbody()

        tm.append_separator_line()
        tm.append_row([" DR", "Peak", "RMS", "Duration", "File name"], 'h')
        tm.append_separator_line()

        for i in range(len(drm.res_list)):

            if drm.res_list[i]['dr14'] > dr14.min_dr():
                row = []
                row.append(" DR%d" % drm.res_list[i]['dr14'])
                row.append(" %.2f" % drm.res_list[i]['dB_peak'] + ' dB')
                row.append(" %.2f" % drm.res_list[i]['dB_rms'] + ' dB')
                row.append(" %s" % drm.res_list[i]['duration'])
                row.append(" %s" % drm.res_list[i]['file_name'])

                tm.append_row(row)

        tm.end_tbody()

        tm.new_foot()
        tm.append_separator_line()

        tm.add_title("Number of files:    " + str(len(drm.res_list)))
        tm.add_title("Official DR value:  DR%d" % int(drm.dr14))

        tm.append_empty_line()
        tm.add_title(dr14.get_name_version())

        tm.append_closing_line()
        tm.end_foot()

        tm.end_table()

        return tm.write_table()


class WriteDrExtended(WriteDr):

    def __init__(self):
        WriteDr.__init__(self)

    def write_dr(self, drm, tm):

        (head, album_dir) = os.path.split(drm.dir_name)

        tm.new_table()

        tm.new_head()
        tm.append_separator_line()

        album_t = drm.meta_data.get_album_title()
        artist = drm.meta_data.get_album_artist()[0]

        if not isinstance(tm, table.TextTable):
            self.set_loudness_war_db_compatible(False)

        if self.get_loudness_war_db_compatible():

            title = ""

            if album_t == None:
                title = " Analyzed folder:  " + album_dir
            else:
                title = " Analyzed: " + album_t
                if artist != None:
                    title = title + " /  Artist: " + artist
            tm.add_title(title)

        else:

            if album_t == None:
                tm.add_title(" Analyzed folder:  " + album_dir)
            else:
                tm.add_title(" Album: " + album_t)
                if artist != None:
                    tm.add_title(" Artist: " + artist)

        tm.end_head()

        tm.new_tbody()

        tm.append_separator_line()
        tm.append_row(["DR", "Peak", "RMS", "Duration", "Title [codec]"], 'h')
        tm.append_separator_line()

        list_bit = []
        sampl_rate = []

        sum_kbs = 0
        cnt = 0

        d_nr = 0

        for i in range(len(drm.res_list)):

            if drm.res_list[i]['dr14'] > dr14.min_dr():
                row = []
                row.append(" DR%d" % drm.res_list[i]['dr14'])
                row.append(" %.2f" % drm.res_list[i]['dB_peak'] + ' dB')
                row.append(" %.2f" % drm.res_list[i]['dB_rms'] + ' dB')
                row.append(drm.res_list[i]['duration'])

                #print( "> " + drm.res_list[i]['file_name'] )

                curr_file_name = drm.res_list[i]['file_name']

                tr_title = drm.meta_data.get_value(curr_file_name, 'title')
                #print( "> " + tr_title )
                if tr_title == None:
                    row.append(drm.res_list[i]['file_name'])
                else:
                    nr = drm.meta_data.get_value(curr_file_name, 'track_nr')
                    codec = drm.meta_data.get_value(curr_file_name, 'codec')

                    if nr == None:
                        nr = i + 1

                    row.append("%02d - %s \t [%s]" % (nr, tr_title, codec))

                bitrate = drm.meta_data.get_value(curr_file_name, 'bitrate')
                bit = drm.meta_data.get_value(curr_file_name, 'bit')
                s_rate = drm.meta_data.get_value(
                    curr_file_name, 'sampling_rate')

                kbs = drm.meta_data.get_value(curr_file_name, 'bitrate')

                if kbs != None:
                    sum_kbs += int(kbs)
                    cnt = cnt + 1

                list_bit.append(bit)
                sampl_rate.append(s_rate)

                tm.append_row(row)

        tm.end_tbody()

        tm.new_foot()
        tm.append_separator_line()

        tm.add_title(" Number of files:    " + str(len(drm.res_list)))
        tm.add_title(" Official DR value:  DR%d" % int(drm.dr14))

        tm.append_empty_line()

        t = set(sampl_rate)
        if len(t) > 1:
            tm.add_title(f" Sampling rate: \t\t various - {sorted(t)} Hz")
        else:
            tm.add_title(f" Sampling rate: \t\t {t.pop()} Hz")

        if cnt > 0:
            tm.add_title(" Average bitrate: \t\t %d kbs " %
                         ((sum_kbs / 1000) / cnt))

        t = set(list_bit)
        if len(t) > 1:
            tm.add_title(f" Bits per sample: \t\t various - {sorted(t)} bit")
        else:
            tm.add_title(f" Bits per sample: \t\t {t.pop()} bit")

        tm.append_empty_line()
        tm.add_title(dr14.get_name_version())

        tm.append_closing_line()
        tm.end_foot()

        tm.end_table()

        return tm.write_table()



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import os
import sys
import tempfile
import subprocess
import re
import wave
import numpy


from dr14meter.out_messages import print_msg, dr14_log_info
from dr14meter.dr14_global import get_ffmpeg_cmd

#from _ftdi1 import NONE

# ffmpeg -i example.m4a -f wav pipe:1 > test.wav

# def check_command(cmd):
#     try:
#         subprocess.run([cmd, '--version'], check=True, stdout=subprocess.DEVNULL)
#         return True
#     except FileNotFoundError:
#         return False


class AudioFileReader:

    def __init__(self):
        # self.use_ffmpeg = False
        self.__ffmpeg_cmd = get_ffmpeg_cmd()

        if sys.platform.startswith('win'):
            self.__cmd = ".\\decoder\\%s " % self.get_cmd()
        else:
            self.__cmd = "%s " % self.get_cmd()

    def get_cmd(self):
        return self.get_ffmpeg_cmd()

    def get_ffmpeg_cmd(self):
        return self.__ffmpeg_cmd

    def get_cmd_options(self, file_name, tmp_file):
        return self.get_generic_ffmpeg_options(file_name, tmp_file)

    def to_wav(self, file_name):

        full_command = self.__cmd

        (head, file) = os.path.split(file_name)
        tmp_dir = tempfile.gettempdir()
        tmp_file = os.path.join(tmp_dir, file) + ".wav"

        file_name = re.sub(r"(\"|`)", r"\\\1", file_name)
        tmp_file = re.sub(r"(\"|`)", r"_xyz_", tmp_file)

        full_command = full_command + " " + \
            self.get_cmd_options(file_name, tmp_file)

        r = subprocess.Popen(full_command, shell=True,
                            stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        stdout_data, stderr_data = r.communicate()

        if os.path.exists(tmp_file):
            return tmp_file
        else:
            return ""

    def read_audio_file_new(self, file_name, target):

        time_a = time.time()

        full_command = self.__cmd

        (head, file) = os.path.split(file_name)
        tmp_dir = tempfile.gettempdir()
        tmp_file = os.path.join(tmp_dir, file) + ".wav"

        file_name = re.sub(r"(\"|`|\$)", r"\\\1", file_name)
        tmp_file = re.sub(r"(\"|`|\$)", r"_xyz_", tmp_file)

        full_command = full_command + " " + \
            self.get_cmd_options(file_name, tmp_file)

        #print_msg( full_command )

        r = subprocess.Popen(full_command, shell=True,
                            stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        stdout_data, stderr_data = r.communicate()

        #read_wav.read_wav( tmp_file )

        ret_f = self.read_wav(tmp_file, target)

        if os.path.exists(tmp_file):
            os.remove(tmp_file)
        else:
            print_msg(file_name + ": unsupported encoder")
            print_msg(f'Command: {full_command}\nError: {stderr_data}')

        time_b = time.time()
        dr14_log_info(
            "AudioFileReader.read_audio_file_new: Clock: %2.8f" % (time_b - time_a))

        return ret_f

    def read_wav(self, file_name, target):

        time_a = time.time()

        convert_8_bit = numpy.float32(2**8 + 1.0)
        convert_16_bit = numpy.float32(2**15 + 1.0)
        convert_32_bit = numpy.float32(2**31 + 1.0)

        try:
            wave_read = wave.open(file_name, 'r')
            target.channels = wave_read.getnchannels()
            target.Fs = wave_read.getframerate()
            target.sample_width = wave_read.getsampwidth()

            nframes = wave_read.getnframes()
            #print_msg( file_name + "!!!!!!!!!!!!: " + str(target.channels) + " " + str(target.sample_width ) + " " + str( target.Fs ) + " " + str( nframes ) )

            X = wave_read.readframes(wave_read.getnframes())

            sample_type = "int%d" % (target.sample_width * 8)

            target.Y = numpy.fromstring(X, dtype=sample_type).reshape(
                nframes, target.channels)

            wave_read.close()

            if sample_type == 'int16':
                target.Y = target.Y / (convert_16_bit)
            elif sample_type == 'int32':
                target.Y = target.Y / (convert_32_bit)
            else:
                target.Y = target.Y / (convert_8_bit)

            #print_msg( "target.Y: " + str(target.Y.dtype) )
        except:
            self.__init__()
            print_msg("Unexpected error: %s" % str(sys.exc_info()))
            print_msg("\n - ERROR ! ")
            return False

        time_b = time.time()
        dr14_log_info("AudioFileReader.read_wav: Clock: %2.8f" %
                      (time_b - time_a))

        return True

    def get_generic_ffmpeg_options(self, file_name, tmp_file):
        return " -i \"%s\" -b:a 16 -ar 44100 -y \"%s\" -loglevel quiet " % (file_name, tmp_file)


class Mp3FileReader(AudioFileReader):
    pass
    # def get_cmd(self):
    #     ret = "lame"
    #     if check_command(ret):
    #         return ret
    #     else:
    #         self.use_ffmpeg = True
    #
    # def get_cmd_options(self, file_name, tmp_file):
    #     return "--silent " + "--decode " + "\"" + file_name + "\"" + " \"%s\" " % tmp_file



class OggFileReader(AudioFileReader):

    def get_cmd(self):
        return "oggdec"

    def get_cmd_options(self, file_name, tmp_file):
        return "--quiet " + "\"" + file_name + "\"" + " --output \"%s\"  " % tmp_file


class WavFileReader(AudioFileReader):

    def read_audio_file_new(self, file_name, target):
        return self.read_wav(file_name, target)

    def get_cmd(self):
        return ""

    def get_cmd_options(self, file_name, tmp_file):
        return ""




# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

from dr14meter.table import *  # ExtendedTextTable
from dr14meter.audio_analysis import *
# from dr14meter.dr14_global import *
import dr14meter.dr14_global as dr14
from dr14meter.dr14_config import *  # get_db_path, enable_db, database_exists
from dr14meter.write_dr import WriteDr, WriteDrExtended
from dr14meter.out_messages import *  # print_out, print_err
# from dr14meter.dr14_utils import *  # test_path_validity

from dr14meter.database.database import dr_database_singletone
from dr14meter.database.query import *

import sys
import re


def local_dr_database_configure():

    print_out("---------------------------------------------------------------------------------------------- ")
    print_out(f"- {dr14.get_exe_name()} --  ")
    print_out("- Local DR database - Configuration procedure ")
    print_out("---------------------------------------------------------------------------------------------- ")

    print_out("  ")
    print_out("  The database, if enabled, automatically stores all DR result in a local database ")
    print_out("  - You must set up some parameters to use the database -")
    print_out("  ")

    flag = True
    while flag:
        print_out("  1. Insert the database directory: Default [%s]:" % os.path.split(
            get_db_path())[0])

        if sys.version_info[0] == 2:
            db_path = raw_input("     > ")
        else:
            db_path = input("     > ")

        db_path = os.path.expanduser(db_path)
        db_path = os.path.expandvars(db_path)

        if re.sub(r"\s+", "", db_path) == "":
            flag = False
            db_path = os.path.split(get_db_path())[0]
            continue

        db_path = os.path.abspath(db_path)

        if test_path_validity(db_path):
            flag = False
        else:
            print_out(
                "  - [ %s ] is not a directory, please insert an acceptable directory name" % db_path)

    print_out("  ")
    print_out(
        "---------------------------------------------------------------------------------------------- ")
    print_out("  ")
    print_out(
        "     If you set a collection directory ONLY the tracks inside this base folder and sub-folders ")
    print_out("     will be added to the database.")
    print_out(
        "     If you insert \'any\' all analyzed tracks will be added to the database ")

    flag = True
    while flag:
        print_out(
            "  2. Insert your collection directory: Default [%s] " % "any")

        if sys.version_info[0] == 2:
            coll_path = raw_input("     > ")
        else:
            coll_path = input("     > ")

        coll_path = os.path.expanduser(coll_path)
        coll_path = os.path.expandvars(coll_path)

        if re.sub(r"(\s+)", "", coll_path) in ["", "any"]:
            flag = False
            coll_path = "/"
            continue

        coll_path = os.path.abspath(coll_path)

        if os.path.isdir(coll_path):
            flag = False
        else:
            print_out(
                "  - [ %s ] is not a directory, please insert an existing directory name" % coll_path)

    print_out("  ")
    print_out(" type: %s -q " % dr14.get_exe_name())
    print_out(" For more details and querying the database ")
    print_out("---------------------------------------------------------------------------------------------- ")

    return (db_path, coll_path)


def fix_problematic_database():

    print_out("  ")
    print_out(" It likes that your database have some problems ")
    print_out("  ")
    print_out(" Options: ")
    print_out("   1. Rebuild the database and back-up the damaged db")
    print_out("   2. Disable the database ")
    print_out("  ")
    nr = input_number(p=" > ", rng=(1, 2))

    dbp = get_db_path()

    if nr == 1:
        if os.path.isfile(dbp):
            dest_file = dbp + ".d_save"
            os.rename(dbp, dest_file)
            print_out("  ")
            print_out(
                " The old database has been saved in the file: %s " % dest_file)
            print_out("  ")

        enable_database(prompt=False)

    elif nr == 2:
        enable_db(False)
        print_out(" The database has been disabled ")
        print_out(" Type: %s --enable_database " % dr14.get_exe_name())
        print_out(" For enabling the database ")
        print_out("  ")


def enable_database(prompt=True):

    f = False

    if database_exists() and db_is_enabled():
        f = True

    if database_exists() and not db_is_enabled():
        db = dr_database_singletone().get()
        f = db.is_db_valid()

        if not f:
            dbp = get_db_path()
            dest_file = dbp + ".de_save"
            os.rename(dbp, dest_file)
        else:
            f = True

    if not database_exists():

        if prompt:
            (db_path, coll_dir) = local_dr_database_configure()
        else:
            db_path = os.path.split(get_db_path())[0]
            coll_dir = get_collection_dir()

        try:
            os.makedirs(db_path)
        except:
            pass

        db_path += "/dr14_database.db"

        set_db_path(db_path)
        set_collection_dir(coll_dir)

        print_msg("Preparing database .... ")
        db = dr_database_singletone().get()
        db.build_database()
        f = db.is_db_valid()

    if f:
        enable_db(True)
        print_msg(
            "The local DR database is ready and enabled! It is located in the file: %s  " % get_db_path())
    else:
        print_err("The building procedure of the database has failed ... retry ")
    return


def input_number(p=" > ", rng=(0, 2**31)):

    flag = True

    while flag:
        if sys.version_info[0] == 2:
            nr = raw_input(" > ")
        else:
            nr = input(" > ")

        try:
            nr = int(nr)
        except:
            print_out(" !! Please insert a valid number ")
            continue

        if rng is not None and not (nr >= rng[0] and nr <= rng[1]):
            print_out(" !! Please insert a valid option number in the [%d .. %d]" % (
                min(rng), max(rng)))
            continue

        flag = False

    return nr


def extended_options(opt_nr):

    qa = False
    qb = False

    if opt_nr in [1, 2, 3, 5, 6]:
        question_a = "Insert the number of desired results: "
        qa = True

    if opt_nr in [3]:
        question_b = "Insert the minimal number of tracks required for inserting an artist in the result: "
        qb = True

    res = []

    if qa:
        print_out(question_a)
        res.append(input_number())

    if qb:
        print_out(question_b)
        res.append(input_number())

    return res


def query_helper():
    print_out("----- QUERY HELPER -----")
    print_out("  ")
    print_out(" Choose one of the following query: ")
    print_out("  ")
    print_out(" 1. The list of the best DR tracks ")
    print_out(" 2. The list of the best DR Albums ")
    print_out(" 3. The list of the best DR Artists ")
    print_out(" 4. The DR time evolution, according with your collection")
    print_out(" 5. The list of the worst DR tracks  ")
    print_out(" 6. The list of the worst DR albums ")
    print_out(" 7. The DR histogram ")
    print_out(" 8. Used audio CODEC and mean DR ")
    print_out(" 0. Exit ")
    print_out("  ")

    nr = input_number(rng=(0, 9))

    class options:
        pass

    ext_opt = []

    if nr in [1, 2, 3, 5, 6]:
        print_out(
            " Do you want to set the query extended options? [N/y] (default: No)")
        if sys.version_info[0] == 2:
            an = raw_input(" > ")
        else:
            an = input(" > ")

        if an.lower() in ["y", "yes"]:
            ext_opt = extended_options(nr)

    if nr == 0:
        return
    elif nr == 1:
        options.query = ["top"]
    elif nr == 2:
        options.query = ["top_alb"]
    elif nr == 3:
        options.query = ["top_art"]
    elif nr == 4:
        options.query = ["evol"]
    elif nr == 5:
        options.query = ["worst"]
    elif nr == 6:
        options.query = ["worst_alb"]
    elif nr == 7:
        options.query = ["hist"]
    elif nr == 8:
        options.query = ["codec"]

    for e in ext_opt:
        options.query.append(e)

    eq_cmd = "%s -q " % get_exe_name()
    for o in options.query:
        eq_cmd += "%s " % o

    print_out("")
    print_out("equivalent command line:\n>    %s " % eq_cmd)
    print_out("")

    res = database_exec_query(options, tm=ExtendedTextTable())

    print_out("")
    print_out(res)


def database_exec_query(options, tm=ExtendedTextTable()):

    if len(options.query) >= 2:
        limit = int(options.query[1])
    else:
        limit = 30

    if len(options.query) >= 3:
        second_opt = options.query[2]
    else:
        second_opt = None

    if options.query[0] == "help":
        query_helper()
        return

    elif options.query[0] == "top":
        table_title = "Top DR tracks"
        q = query_top_dr()

    elif options.query[0] == "top_alb":
        table_title = "Top DR Albums"
        q = query_top_albums_dr()

    elif options.query[0] == "worst":
        table_title = "Worst DR Tracks"
        q = query_worst_dr()

    elif options.query[0] == "worst_alb":
        table_title = "Worst DR Albums"
        q = query_worst_albums_dr()

    elif options.query[0] == "top_art":
        table_title = "Top Artists"
        q = query_top_artists()
        if second_opt is not None:
            q.min_track = int(second_opt)

    elif options.query[0] == "hist":
        table_title = "DR histogram"
        q = query_dr_histogram()

    elif options.query[0] == "evol":
        table_title = "DR evolution"
        q = query_date_dr_evolution()

    elif options.query[0] == "codec":
        table_title = "Codec Info"
        q = query_dr_codec()

    else:
        raise Exception("Option not allowed")

    q.limit = limit

    wr = WriteDr()
    table_code = wr.write_query_result(
        q.exec_query(), tm, table_title, q.get_col_keys())

    return table_code



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import argparse


def parse_args():
    desc = "Compute the DR14 value of the audio files according to the algorithm "
    desc = desc + "described by the Pleasurize Music Foundation "
    desc = desc + "Visit: http://www.dynamicrange.de/"

    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument("-1", "--disable_multithread",
                        action="store_true",
                        dest="disable_multithread",
                        help="Disable the multi-Core mode")

    parser.add_argument("-r", "--recursive",
                        action="store_true",
                        dest="recursive",
                        help="Scan recursively the subdirectories")

    parser.add_argument("-a", "--append",
                        action="store_true",
                        dest="append",
                        help="Append all results in a single file; it should be used in couple with -r")

    parser.add_argument("-b", "--basic_table",
                        action="store_true",
                        dest="basic_table",
                        help="Write the resulting tables in the basic format")

    parser.add_argument("-n", "--turn_off_out",
                        action="store_true",
                        dest="turn_off_out",
                        help="do not writes the output files")

    parser.add_argument("-p", "--print_std_out",
                        action="store_true",
                        dest="print_std_out",
                        help="writes the full result on the std_out")

    parser.add_argument("-o", "--outdir",
                        action="store",
                        dest="out_dir",
                        type=str,
                        help="Write the resultings files into the given directory")

    parser.add_argument("-t", "--tables",
                        action="store",
                        choices='htbwa',
                        dest="out_tables",
                        default="t",
                        help="Select the output files to be written. h=html t=text b=bbcode w=mediawiki a=all_formats")

    parser.add_argument("-f", "--file",
                        action='store_true',
                        dest="scan_file",
                        help="Compute the DR14 of a single file and exit")

    parser.add_argument("--enable_database",
                        action="store_true",
                        dest="enable_database",
                        help="Enable and create, if necessary, the local DR database")

    parser.add_argument("--disable_database",
                        action="store_true",
                        dest="disable_database",
                        help="Disable the local DR database")

    parser.add_argument("--dump_database",
                        action="store_true",
                        dest="dump_database",
                        help="Dump the local DR database")

    parser.add_argument("-q", "--query",
                        nargs="*",
                        dest="query",
                        help="""query the database. Options: [-q] [-q top #nr] [-q worst #nr] \n 
                                             [-q top_alb #nr] [-q worst_alb #nr] [-q top_art #nr #mt] 
                                             [-q hist] [-q evol] [-q codec] \n 
                                             [-q help] """ )

    parser.add_argument("-d", "--dr_database",
                        action="store_true",
                        dest="dr_database",
                        help="Output file compatible with the DR database at http:///www.dr.loudness-war.info")

    parser.add_argument("--hist",
                        action="store_true",
                        dest="histogram",
                        help="Plot the histogram of dynamic of a single file and exit")

    parser.add_argument("--lev_hist",
                        action="store_true",
                        dest="lev_histogram",
                        help="Plot the histogram of the saples levels of a single file and exit")

    parser.add_argument("--spectrogram",
                        action="store_true",
                        dest="spectrogram",
                        help="Plot the spectrogram of a single file and exit (beta)")

    parser.add_argument("--plot_track",
                        action="store_true",
                        dest="plot_track",
                        help="Plot the track of the given file and exit")

    parser.add_argument("--plot_track_dst",
                        action="store_true",
                        dest="plot_track_dst",
                        help="Plot the track in a cool manner of the given file and exit")

    parser.add_argument("--compress",
                        action="store",
                        choices=["very_soft", "v", "soft", "s",
                                 "medium", "m", "hard", "h", "very_hard", "vh"],
                        default=None,
                        dest="compress",
                        help="Perform the dynamic compression on a single file and exit, the resulting track will be written in a temp directory")

    parser.add_argument("--dyn_vivacity",
                        action="store_true",
                        dest="dynamic_vivacity",
                        help="Plot the graph of the dynamic vivacity of a single audio file and exit")

    parser.add_argument("--files_list",
                        action='store_true',
                        dest="files_list",
                        help="takes a files list and compute the DR of each file, if no files are specified, it reads from STDIN")

    parser.add_argument("--quiet",
                        action="store_true",
                        dest="quiet",
                        help="Quiet mode")

    # parser.add_argument("--skip_version_check",
    #                     action="store_true",
    #                     dest="skip_version_check",
    #                     help="Do not check for new versions")

    parser.add_argument("-v", "--version",
                        action="store_true",
                        dest="version",
                        help="print the current version and exit")

    parser.add_argument("--tag", 
                        action="store_true",
                        dest="tag",
                        help="Write the DR to the tag of the input files")

    parser.add_argument(
        dest="path_name",
        nargs='?',
        default=None
    )

    return parser.parse_args()



# dr14meter: compute the DR14 value of the given audio files
# Copyright (C) 2024  pe7ro
#
# dr14_t.meter: compute the DR14 value of the given audiofiles
# Copyright (C) 2011  Simone Riva
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from numpy import *
from dr14meter.audio_math import *
import numpy


def compute_DRV(Y, Fs, duration=None, Dr_lr=None):

    s = Y.shape

    if len(Y.shape) > 1:
        ch = s[1]
    else:
        ch = 1

    block_time = 1
    block_samples = block_time * (Fs)

    threshold = 0.15
    seg_cnt = int(floor(s[0] / block_samples))

    if seg_cnt < 3:
        return (0, -100, -100)

    curr_sam = 0

    rms = zeros((seg_cnt, ch))
    peaks = zeros((seg_cnt, ch))

    for i in range(seg_cnt - 1):
        r = arange(curr_sam, curr_sam + block_samples)

        rms[i, :] = decibel_u(u_rms(Y[r, :]), 1.0)
        peaks[i, :] = decibel_u(numpy.max(numpy.abs(Y[r, :])), 1.0)

        curr_sam = curr_sam + block_samples

    i = numpy.logical_or(rms == numpy.nan, rms == -numpy.inf)
    rms[i] = decibel_u(audio_min(), 1.0)

    i = numpy.logical_or(peaks == numpy.nan, peaks == -numpy.inf)
    peaks[i] = decibel_u(audio_min(), 1.0)

    Ydr = numpy.mean(peaks - rms, 1)

    (n, bins) = numpy.histogram(Ydr, 100)

    #print( peaks )
    #print( rms )

    max_freq = numpy.max(n)

    bs = bins.shape[0]
    bins = bins[0:bs - 1] + numpy.diff(bins) / 2.0

    i = n > max_freq * threshold

    n = n[i]
    bins = bins[i]

    m = numpy.sum(n * bins) / numpy.sum(n)

    drV = round(m - 3)

    dB_peak = numpy.max(peaks)

    dB_rms = numpy.sum(numpy.mean(rms, 0)) / 2.0

    if duration != None:
        duration.tm_min = int(s[0] * (1.0 / Fs) / 60.0)
        duration.tm_sec = int(
            ((s[0] * (1.0 / Fs) / 60.0) - duration.tm_min) * 60.0)

    if Dr_lr != None:
        Dr_lr = ch_dr14

    return (drV, dB_peak, dB_rms)
